#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
This script calls SPADES assembler and plasmidSPADES. It also generates descriptive statistics of the assembly process.

This script can be called as a single script if desired or called from the BacterialTyper module assemble or as a python api.
"""
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
from io import open
from Bio import SeqIO
import shutil
from termcolor import colored

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts.blast_parser import parse
from BacterialTyper.other_tools import tools

################################################
def run_SPADES_plasmid_assembly(path, file1, file2, sample, SPADES_bin, threads):
	"""Generate plasmid assembly using SPADES
	
	- Calls SPADES to assemble plasmids using --plasmid option (using :func:`BacterialTyper.scripts.spades_assembler.SPADES_systemCall`) 
	
	- SPADES generates a file named as *scaffolds.fasta* within the directory provided. This function retrieves path to contigs/scaffolds assembled (using :func:`BacterialTyper.scripts.functions.retrieve_matching_files`).
	
	:param path: Absolute path to folder.
	:param file1: Absolute path to fastq reads (R1).
	:param file2: Absolute path to fastq reads (R2).
	:param sample: Sample name or tag to identify sample
	:param SPADES_bin: Binary executable for SPADES assembly software.
	:param threads: Number of CPUs to use.
	:type path: string
	:type file1: string
	:type file2: string
	:type name: string
	:type threads: integer
	:return: Plasmid contigs/scaffolds assembled.
	:rtype: string : Path to assembly fasta file.
	:warnings: Returns **FAIL** if assembly process stopped.
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.spades_assembler.SPADES_systemCall`
	
		- :func:`BacterialTyper.scripts.functions.retrieve_matching_files`
	"""
	print ('+ Running plasmid assembly...')
	name = sample + '_plasmid'
	options = '--plasmid '
	message_return = SPADES_systemCall(path, file1, file2, name, SPADES_bin, options, threads)

	if 	message_return == 'FAIL':	
		print ("\n\n***ERROR: plasmidSPADES failed for sample " + sample)	
		exit()

	scaffolds_retrieved = functions.retrieve_matching_files(path + '/' + name, "scaffolds.fasta")
	if scaffolds_retrieved == '':	
		print ('\n\n***ATTENTION: No plasmids assembly...')

	return (scaffolds_retrieved[0])

################################################
def run_SPADES_assembly(path, file1, file2, sample, SPADES_bin, threads):
	"""Generate main assembly using SPADES
	
	- Calls SPADES to assemble reads (using :func:`BacterialTyper.scripts.spades_assembler.SPADES_systemCall`) 
	
	- SPADES generates a file named as *scaffolds.fasta* within the directory provided. This function retrieves path to contigs/scaffolds assembled (using :func:`BacterialTyper.scripts.functions.retrieve_matching_files`).
	
	- Renames contigs retrieved using sample name (using :func:`BacterialTyper.scripts.spades_assembler.rename_contigs`).
		
	:param path: Absolute path to folder.
	:param file1: Absolute path to fastq reads (R1).
	:param file2: Absolute path to fastq reads (R2).
	:param sample: Sample name or tag to identify sample
	:param SPADES_bin: Binary executable for SPADES assembly software.
	:param threads: Number of CPUs to use.
	:type path: string
	:type file1: string
	:type file2: string
	:type name: string
	:type threads: integer
	:return: Contigs/scaffolds assembled renamed.
	:rtype: string : Path to assembly fasta file.
	:warnings: Returns **FAIL** if assembly process stopped.
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.spades_assembler.SPADES_systemCall`
	
		- :func:`BacterialTyper.scripts.functions.retrieve_matching_files`
	
		- :func:`BacterialTyper.scripts.functions.rename_fasta_seqs`
	"""
	##print ('+ Running main assembly...')
	options = ''
	message_return = SPADES_systemCall(path, file1, file2, sample, SPADES_bin, options, threads)
	if 	message_return == 'FAIL':	
		print ("\n\n***ERROR: SPADES failed for sample " + sample)
		return ('FAIL')

	scaffolds_retrieved = functions.retrieve_matching_files(path, "scaffolds.fasta")
	if scaffolds_retrieved == '':	
		print ('\n\n***ERROR: No scaffolds assembly...')
		return ('FAIL')
	
	### Due to limiations with Genbank format, no more thatn 37 characters are supported for 
	### locus tag identification. This might affect later annotation process and subsequent analysis
	### https://github.com/tseemann/prokka/issues/337 
	new_contigs = path + '/' + sample + '_assembly.fna'
	id_conversion_file = functions.rename_fasta_seqs(scaffolds_retrieved[0], sample, new_contigs)
		
	if	id_conversion_file == 'FAIL':	
		print ("\n\n***ERROR: Rename contigs failed for sample " + sample)
		return ('FAIL')
	else:
		print ("+ Name conversion details saved in file " + id_conversion_file)
	
	return (new_contigs)

################################################
def SPADES_systemCall(sample_folder, file1, file2, name, SPADES_bin, options, threads):
	"""Generate SPADES system call.
	
	It calls system for SPADES and generates time stamp file in the folder provided (sample_folder + '/.success_assembly') for later analysis.
	
	Steps:
	
	- It generates system call for SPADES assembly (using :func:`BacterialTyper.scripts.functions.system_call`). 
	
	- It generates timestamp file (See :func:`BacterialTyper.scripts.functions.print_time_stamp`).
	
	:param sample_folder: Absolute path to store results. It must exists.
	:param file1: Absolute path to fastq reads (R1).
	:param file2: Absolute path to fastq reads (R2).
	:param name: Sample name or tag to identify sample.
	:param SPADES_bin: Binary executable for SPADES assembly software.
	:param options: Plasmid assembly is possible if specificed via options (--plasmid).
	:param threads: Number of CPUs to use.
	
	:type name: string
	:type sample_folder: string
	:type file1: string
	:type file2: string
	:type SPADES_bin: string
	:type options: string
	:type threads: integer
	
	:return: Returns **OK** if assembly process succeeded and fasta file is generated.
	:rtype: string.
	:warnings: Returns **FAIL** if assembly process stopped.
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.system_call`
	
		- :func:`BacterialTyper.scripts.functions.print_time_stamp`
	"""
	
	## check if previously assembled and succeeded
	filename_stamp = sample_folder + '/.success_assembly'
	if os.path.isfile(filename_stamp):
		stamp =	functions.read_time_stamp(filename_stamp)
		print (colored("\tA previous command generated results on: %s [%s]" %(stamp, name), 'yellow'))
		return('OK')

	## call system for SPADES sample given
	logFile = sample_folder + '/' + name + '.log'
	
	## command	
	cmd_SPADES = '%s %s-t %s -o %s -1 %s -2 %s > %s 2> %s' %(SPADES_bin, options, threads, sample_folder, file1, file2, logFile, logFile)
	code = functions.system_call(cmd_SPADES)
	
	if (code == 'OK'):
		## success stamps
		filename_stamp = sample_folder + '/.success_assembly'
		stamp =	functions.print_time_stamp(filename_stamp)
		return('OK')

	return "FAIL"
	
################################################
def run_module_assembly(name, folder, file1, file2, threads):
	"""Assembly main module call.
	
	It calls assembly function to process data provided and returns genome statistics. Steps: 
	
	- Retrieves SPADES_ executable (See details :func:`BacterialTyper.scripts.set_config.get_exe`) using the minimun version required (See :func:`BacterialTyper.scripts.set_config.min_version_programs` for details)
	
	- It generates a call to SPADES_ assembler (See :func:`BacterialTyper.scripts.spades_assembler.run_SPADES_assembly`). 
		
	- If assembly succeeds and fasta file is generated under the directory provided, contig statistics are generated (:func:`BacterialTyper.scripts.spades_assembler.contig_stats`).
	
	- It retrieves spades executable using 
	
	:param name: Sample name or tag to identify sample
	:param folder: Absolute path to folder.
	:param file1: Absolute path to fastq reads (R1).
	:param file2: Absolute path to fastq reads (R2).
	:param threads: Number of CPUs to use.
	:type name: string
	:type folder: string
	:type file1: string
	:type file2: string
	:type threads: integer
	:return: Assembly statistics file.
	:rtype: string : Path to file assembly statistics file.
	:warnings: Returns **FAIL** if assembly process stopped.
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.set_config.get_exe`
	
		- :func:`BacterialTyper.scripts.spades_assembler.run_SPADES_assembly`
	
		- :func:`BacterialTyper.scripts.set_config.min_version_programs`
	
		- :func:`BacterialTyper.scripts.spades_assembler.contig_stats`
	
	.. include:: ../../links.inc	 	
	"""
	
	print ("+ Calling spades assembly for sample...", name)	
	
	## get configuration
	SPADES_bin = set_config.get_exe('spades')
	
	## assembly main 
	path_to_contigs = run_SPADES_assembly(folder, file1, file2, name, SPADES_bin, threads)
	
	if path_to_contigs == 'FAIL':
		return ('FAIL')
	else:
		## contig stats
		#print ('+ Get assembly statistics:...\n')
		contig_out = contig_stats(path_to_contigs, "1000,10000")	
	
		## check statistics in file
		print ("+ Check statistics for sample %s in file:\n%s" %(name, contig_out))
		return(contig_out)

################################################
def discardPlasmids(contigs, plasmids, path, sample):
	
	## check if any plasmids
	if (plasmids == 'FAIL'):
		#print ('+ No plasmids assembled.')
		#print ('+ No need to discard any plasmids from the main assembly')
		
		contig_out_file = os.path.dirname(path) + '/' + sample + '/' + sample + '_chromosome.fna.tmp'
		shutil.copy(contigs, contig_out_file)
		return (contig_out_file, plasmids)
	
	## discard 
	print ('+ Check if any plasmids are also reported in main assembly...')

	folder = functions.create_subfolder('blast_search', path)	
	
	## makeblastDB
	dbName = folder + '/mainAssembly'
	functions.makeblastdb(dbName, contigs)
	
	## blastn command
	outFile = folder + '/blastn_output.txt'
	threads = 1
	functions.blastn(outFile, dbName, plasmids, threads)
	
	########################
	## parseBlast results
	########################
	
	## thresholds
	eval_thresh_float = float(1e-20)
	aln_thresh_given = 90
	min_length = 1000
	
	outFile_parsed = folder + '/blastn_output_parsed.txt'
	output_file = open(outFile_parsed, 'w')	
	sequences2discard = []

	print ('+ Parsing BLAST results generated...\n')
	## get results
	fh = open(outFile)
	for blast_record in parse(fh, eval_thresh=eval_thresh_float, aln_thresh=aln_thresh_given, length_thresh=min_length):
		for hit in blast_record.hits:
			for hsp in hit:
				output_file.write('****Alignment****')
				output_file.write('\n')
				
				output_file.write('query id: {}'.format(blast_record.qid))
				output_file.write('\n')
				
				sequences2discard.append(hsp.sid)
				output_file.write('sequence: %s' %hsp.sid)
				output_file.write('\n')
				
				output_file.write('e value: %s' %hsp.evalue)
				output_file.write('\n')
				
				output_file.write('aln: %s' %hsp.length)
				output_file.write('\n')
				
				output_file.write('qlen: %s [>%s]' %(hsp.qlen, min_length))
				output_file.write('\n')
				
				aln = (int(hsp.qlen)/int(hsp.slen))*100
				output_file.write('aln/slen: %s [> %s]' %(aln, aln_thresh_given))
				output_file.write('\n\n')

	fh.close()
	output_file.close()
	
	items = len(sequences2discard)
	print ('There are %s sequences to discard from main assembly identified as plasmids' %items)

	## print filtered contigs
	contig_out_file = os.path.dirname(path) + '/' + sample + '/' + sample + '_chromosome.fna.tmp'
	plasmid_out_file = os.path.dirname(path) + '/' + sample + '/' + sample + '_plasmid.fna.tmp'
		
	contig_out_file_handle = open(contig_out_file, 'w')
	plasmid_out_file_handle = open(plasmid_out_file, 'w')
	
	contig_items = SeqIO.parse(contigs, 'fasta')
	for seq in contig_items:
		if seq.id in sequences2discard:
			plasmid_out_file_handle.write(seq.format("fasta"))
			plasmid_out_file_handle.write('\n')
		else:
			contig_out_file_handle.write(seq.format("fasta"))
			contig_out_file_handle.write('\n')

	contig_out_file_handle.close()
	plasmid_out_file_handle.close()	
	
	return (contig_out_file, plasmid_out_file)

################################################
def contig_stats(assembly_file, csv_arg):
	"""Generate assembly statistics
	
	Calls additional perl script to generate contig statistics. Steps:
	
	- Retrieve information of additional perl script location (using :func:`BacterialTyper.other_tools.tools.perl_scripts`)
	
	- Create system call (using :func:`BacterialTyper.scripts.functions.system_call`) and return output statistic file created.
	
	:param assembly_file: Absolute path to assembly fasta file.
	:type assembly_file: string
	:param csv_arg: Comma separated values for splitting sets. Default: 1000,10000
	:type csv_arg: string
	:return: Text file containing statistics for later analysis.
	:rtype: string
	
	The perl script contig_stats.pl has a mandatory argument which is a single fasta file and default splitting sets are: 1000, 10000. User can provide new parts using a csv argument for the script:
	
	.. code-block:: sh

		perl BacterialTyper/other_tools/perl/contig_stats.pl fasta_file 1000,10000
		

	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.system_call`
		
		- :func:`BacterialTyper.other_tools.tools.perl_scripts`
	"""
	contig_stats_script = tools.perl_scripts('contig_stats')

	
	file_out = assembly_file + '_stats.txt'
	cmd_stats = 'perl %s %s %s > %s' %(contig_stats_script, assembly_file, csv_arg, file_out) ## [TODO] Generate this code in python
	code_chr = functions.system_call(cmd_stats)
	return (file_out)

################################################
def	help_options():
	"""Help options when run spades_assembler.py as a single script
	
	Parameters reflected here refer to spades_assembler.py main function.
	
	:param file1: Absolute path to fastq reads (R1).
	:param file2: Absolute path to fastq reads (R2).
	:param sample: Sample name or tag to identify sample
	:param SPADES_bin: Binary executable for SPADES assembly software.
	:param threads: Number of CPUs to use.
	:param path: Absolute path to folder.
	
	:type file1: string
	:type file2: string
	:type sample: string
	:type SPADES_bin: string
	:type threads: integer
	:type path: string	
	"""
	print ("\nUSAGE: python %s file1 file2 name SPADES_bin threads path\n"  %os.path.realpath(__file__))

################################################
def main():
	## control if options provided or help
	if len(sys.argv) > 1:
		print ("")
	else:
		help_options()
		exit()    	

	file1 = os.path.abspath(argv[1])
	file2 = os.path.abspath(argv[2])
	sample = argv[3]
	SPADES_bin = argv[4]
	threads = int(argv[5])
	path = 	argv[6]

	folder = functions.create_subfolder(sample, path)

	## assembly main 
	path_to_contigs = run_SPADES_assembly(folder, file1, file2, sample, SPADES_bin, threads)
	
	## assembly plasmids
	path_to_plasmids = run_SPADES_plasmid_assembly(folder, file1, file2, sample, SPADES_bin, threads)

	## discard plasmids from main
	tmp_contigs, tmp_plasmids = discardPlasmids(path_to_contigs, path_to_plasmids, folder, sample)
	
	## rename fasta sequences
	new_contigs_list = tmp_contigs.split(".tmp")
	new_contigs = new_contigs_list[0]
	rename_contigs(tmp_contigs, "scaffolds_chr", new_contigs)
	
	new_plasmids=""
	if os.path.isfile(tmp_plasmids):
		new_plasmids_list = tmp_plasmids.split(".tmp")
		new_plasmids = new_plasmids_list[0]
		rename_contigs(tmp_plasmids, "scaffolds_plasmids", new_plasmids)
	
	
	## generate contig statistics
	print ('+ Get assembly statistics:...\n')

	## get contig statistics	
	contig_out = contig_stats(new_contigs, "1000,10000")	
	contig_out_file = open(contig_out, 'r')
	contig_out_file_read = contig_out_file.read()
	contig_out_file.close()
	
	## dump in screen
	print (contig_out_file_read)
	print ()	

	if (new_plasmids == 'FAIL'):
		print ('+ No plasmids identified...\n')
	else:
		print ('+ Plasmids assembly')
		plasmid_out = contig_stats(new_plasmids, "1000,10000")	

		## dump in screen
		plasmid_out_file = open(plasmid_out, 'r')
		plasmid_file_read = plasmid_out_file.read()
		plasmid_out_file.close()
		print(plasmid_file_read)	
	
######

'''******************************************'''
if __name__== "__main__":
	main()

################################################
def run_module_SPADES_old(name, folder, file1, file2, threads):

	print ("+ Calling spades assembly for sample...", name)	

	## folder create
	functions.create_folder(folder)
	
	## get configuration
	SPADES_bin = set_config.get_exe('spades')
	
	## assembly main 
	path_to_contigs = run_SPADES_assembly(folder, file1, file2, name, SPADES_bin, threads)

	## assembly plasmids
	path_to_plasmids = run_SPADES_plasmid_assembly(folder, file1, file2, name, SPADES_bin, threads)
	
	## discard plasmids from main
	(tmp_contigs, tmp_plasmids) = discardPlasmids(path_to_contigs, path_to_plasmids, folder, name)
	
	## rename fasta sequences
	new_contigs = tmp_contigs.split(".fna.tmp")[0] + '.fna'	
	rename_contigs(tmp_contigs, "scaffolds_chr", new_contigs)
	
	new_plasmids=""
	if os.path.isfile(tmp_plasmids):
		new_plasmids = tmp_plasmids.split(".fna.tmp")[0] + '.fna'	
		rename_contigs(tmp_plasmids, "scaffolds_plasmids", new_plasmids)
	
	## contig stats
	stats(new_contigs, new_plasmids)
	
	## success stamps
	filename_stamp = folder + '/.success'
	stamp =	functions.print_time_stamp(filename_stamp)




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Clusters fasta files or fastq reads. Using: project data, genbank entries from database or previous samples. 
'''
## import useful modules
import os
import sys
import re
import time
from io import open
import shutil
import concurrent.futures
from termcolor import colored
import pandas as pd

## import my modules
from HCGB import sampleParser
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.modules import help_info
from BacterialTyper.scripts import database_generator
from BacterialTyper.scripts import min_hash_caller

##############################################
def run_cluster(options):

	## init time
	start_time_total = time.time()

	##################################
	### show help messages if desired	
	##################################
	if (options.help_project):
		## information for project
		help_info.project_help()
		exit()
	elif (options.help_Mash):
		## information for Min Hash Software
		min_hash_caller.helpMash()
		exit()
		
	## debugging messages
	global Debug
	if (options.debug):
		Debug = True
	else:
		Debug = False
		
	### set as default paired_end mode
	if (options.single_end):
		options.pair = False
	else:
		options.pair = True
	
	functions.pipeline_header()
	functions.boxymcboxface("Clustering samples")
	print ("--------- Starting Process ---------")
	functions.print_time()

	## absolute path for in & out
	input_dir = os.path.abspath(options.input)
	outdir=""

	## Project mode as default
	project_mode=True
	if (options.detached):
		options.project = False
		project_mode=False
		outdir = os.path.abspath(options.output_folder)
	else:
		options.project = True
		outdir = input_dir	
	
	## get files
	if options.reads:
		if options.noTrim:
			## raw reads
			pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "fastq", ("fastq", "fq", "fastq.gz", "fq.gz"), options.debug)
		else:
			## trimm reads
			pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "trim", ['_trim'], options.debug)

		## keep only R1 reads if paired-end
		if options.pair:
			pd_samples_retrieved = pd_samples_retrieved.loc[pd_samples_retrieved['read_pair']== "R1"]

	else:
		## default
		pd_samples_retrieved = sampleParser.files.get_files(options, input_dir, "assembly", ["fna"], options.debug)

	## debug message
	if (Debug):
		print (colored("**DEBUG: pd_samples_retrieve **", 'yellow'))
		print (pd_samples_retrieved)
	
	# exit if empty
	if pd_samples_retrieved.empty:
		print ("No data has been retrieved from the project folder provided. Exiting now...")
		exit()


	## generate output folder, if necessary
	print ("\n+ Create output folder(s):")
	if not options.project:
		functions.create_folder(outdir)
	
	## for each sample
	outdir_dict = functions.outdir_project(outdir, options.project, pd_samples_retrieved, "mash")	

	## debug message
	if (Debug):
		print (colored("**DEBUG: outdir_dict **", 'yellow'))
		print (outdir_dict)

	## get databases to check
	retrieve_databases = get_options_db(options)
	
	## time stamp
	start_time_partial = functions.timestamp(start_time_total)
	
	## remove samples if specified
	if options.ex_sample:
		ex_samples = functions.get_info_file(options.ex_sample)
		retrieve_databases = retrieve_databases.loc[~retrieve_databases.index.isin(ex_samples)]
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: retrieve_database **", 'yellow'))
		pd.set_option('display.max_colwidth', None)
		pd.set_option('display.max_columns', None)
		print (retrieve_databases)

	## check if all samples in user_data or genbank are indexed
	siglist_all = []
	for index, row in retrieve_databases.iterrows():
		if not row['path'] == 'NaN':
			if (Debug):
				functions.print_sepLine("*",25, False)
				print (row)
				
			if all([ int(options.kmer_size) == int(row['ksize']), int(options.n_sketch) == int(row['num_sketch']) ]):
				siglist_all.append(min_hash_caller.read_signature(row['path'], options.kmer_size))
				continue

		## index assembly or reads...
		(sigfile, siglist) = generate_sketch(row['folder'], row['original'], index, options.kmer_size, options.n_sketch, Debug)
		retrieve_databases.loc[index]['path'] = sigfile
		retrieve_databases.loc[index]['ksize'] = options.kmer_size
		retrieve_databases.loc[index]['num_sketch'] = options.n_sketch
		siglist_all.append(siglist)
	
	### Cluster project samples
	print (colored("\n+ Collect project data", 'green'))
	print ("+ Generate mash sketches for each sample analyzed...")
	pd_samples_retrieved = pd_samples_retrieved.set_index('name')
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: pd_samples_retrieved **", 'yellow'))
		print (pd_samples_retrieved)

	## init dataframe for project data
	colname = ["source", "name", "path", "original", "ksize", "num_sketch"]
	pd_samples_sketched  = pd.DataFrame(columns = colname)
	for index, row in pd_samples_retrieved.iterrows():
		if index in retrieve_databases.index:
			print (colored('\t+ Sketched signature (%s) available within user data...' %index, 'yellow'))
			continue

		this_sig = outdir_dict[index] + '/' + index + '.sig'
		if os.path.exists(this_sig):
			## File signature might exist

			## read original
			file2print = outdir_dict[index] + '/.original'
			if not os.path.exists(file2print):
				original = ['NaN']
			else:
				original = functions.readList_fromFile(file2print)
				if all([ int(options.kmer_size) == int(original[1]), int(options.n_sketch) == int(original[2])]):
					siglist_all.append(min_hash_caller.read_signature(this_sig, options.kmer_size)) 
					pd_samples_sketched.loc[len(pd_samples_sketched)] = ('project_data', index, this_sig, row['sample'], options.kmer_size, options.n_sketch)
					print (colored('\t+ Sketched signature available (%s) in project folder...' %index, 'green'))
					continue			
					
		print (colored('\t+ Sketched signature to be generated: (%s)...' %index, 'yellow'))
		## index assembly or reads...
		(sigfile, siglist) = generate_sketch(outdir_dict[index], row['sample'], index, options.kmer_size, options.n_sketch, Debug)
		pd_samples_sketched.loc[len(pd_samples_sketched)] = ('project_data', index, sigfile, row['sample'], options.kmer_size, options.n_sketch)
		siglist_all.append(siglist)

	print ("\n+ Clustering sequences...")
	pd_samples_sketched = pd_samples_sketched.set_index('name')
	
	####
	if retrieve_databases.empty: 
		cluster_df = pd_samples_sketched
	else:
		tmp = retrieve_databases[['source', 'db', 'path', 'original', 'ksize', 'num_sketch']]
		tmp = tmp.rename(columns={'db': 'name'})
		tmp.set_index('name')
		
		if (Debug):
			print (colored("**DEBUG: tmp **", 'yellow'))
			print (tmp)	
		
		## merge both dataframes
		cluster_df = pd.concat([pd_samples_sketched, tmp], join='inner', sort=True)
	
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: pd_samples_sketched **", 'yellow'))
		print (pd_samples_sketched)
		
		print (colored("**DEBUG: cluster_df **", 'yellow'))
		print (cluster_df)
		
		print (colored("**DEBUG: Signatures **", 'yellow'))
		print (siglist_all)
		
		print (colored("**DEBUG: length siglist_all **", 'yellow'))
		print (len(siglist_all))

	## Assign Colors colorLabels	
	color_df = cluster_df.filter(["source"], axis=1)
	color_df["color"] = "r" ## red::genbank

	## project data
	project_data = list( color_df[ color_df["source"] == "project_data"].index)
	color_df.loc[ color_df.index.isin(project_data), "color"] = "g" ## green::project_data

	## user_data
	user_data = list( color_df[ color_df["source"] == "user_data"].index)
	color_df.loc[color_df.index.isin(user_data), "color"] = "b" ## blue::user_data
	
	colorLabels = color_df['color'].to_dict()
	
	if Debug:
		print(color_df)
		print(colorLabels)

	## parse results
	if options.project:
		outdir_report = functions.create_subfolder("report", outdir)
		#final_dir = outdir + '/report/cluster'
		final_dir = functions.create_subfolder("cluster", outdir_report) 
	else:
		final_dir = outdir

	## compare
	name = 'cluster_' + str(functions.create_human_timestamp())
	tag_cluster_info = final_dir + '/' + name
	print ('+ Saving results in folder: ', final_dir)
	print ('\tFile name: ', name)
	(DataMatrix, labeltext) = min_hash_caller.compare(siglist_all, tag_cluster_info, Debug)

	## get colorLabels	
	
	## plot images
	pdf = True
	cluster_returned = min_hash_caller.plot(DataMatrix, labeltext, tag_cluster_info, pdf, colorLabels)
	
	## generate newick tree 
	min_hash_caller.get_Newick_tree(cluster_returned, DataMatrix, labeltext, tag_cluster_info)	
	
	return()

############################################################	
def get_options_db(options):
	##
	## Among all databases available and according to the input options,
	## select the databases to use and set dataframe with this information
	##
	
	print ("\n\n+ Select databases to use for identification:")
	
	### database folder to use
	database2use = os.path.abspath(options.database)
	
	## debug message
	if (Debug):
		print (colored("**DEBUG: Database to use: " +  database2use + " **", 'yellow'))
	
	## external file provided: single or batch
	if (options.external_file):
		abs_path_ext_file = os.path.abspath(options.external_file)
		if options.batch_external:
			myList = functions.readList_fromFile(abs_path_ext_file)
			join_str = ','.join(myList)
		else:
			join_str = abs_path_ext_file

	############################################################
	### Options: according to user input: select databases to use
	option_db = ""
	
	############
	## 1) only user data: previously identified and added
	############
	if (options.user_data):
		option_db = "Mash:user_data"
		
	############
	## 2) only genbank data: previously download from NCBI reference genomes
	############
	elif (options.genbank_data):
		option_db = "Mash:genbank"
	
	############
	## 3) only project_data
	############
	elif (options.only_project_data):
		option_db = "Mash:project_data"
		pd_MASH = pd.DataFrame()
		return(pd_MASH)

	############
	## 4) only external_data
	############
	elif (options.only_external_data):
		option_db = "Mash_external_data:" + join_str

	#################
	## all databases 
	#################
	else:		
		#############################
		option_db = 'Mash:user_data#Mash:genbank'
		if (options.external_file):
			option_db = option_db + '#Mash_external_data:' + join_str
	
	###############
	### get dbs
	###############
	print ("\n+ Parsing information to retrieve databases")
	print ("+ Reading from database: " + database2use)
	functions.print_sepLine("-",50, False)

	###############
	## debug message
	if (Debug):
		print (colored("**DEBUG: option_db: " +  option_db + " **", 'yellow'))

	pd_MASH = database_generator.getdbs("MASH", database2use, option_db, Debug)
	functions.print_sepLine("-",50, False)

	## return both dataframes
	return (pd_MASH)

############################################################	
def generate_sketch(folder, assembly, entry, ksize, n_sketch, Debug):

	(sigfile, siglist) = min_hash_caller.sketch_database({ entry: assembly }, folder, Debug, ksize, n_sketch)
	#functions.print_sepLine("*",50, False)
	
	## print original in file
	file2print = folder + '/.original'
	
	## do not write full path because database can move.
	## as long as mantains the folder organization it should work.
	assembly_tmp_path = "../assembly/" + os.path.basename(assembly)
	
	list_fna = [assembly_tmp_path, str(ksize), str(n_sketch)]
	functions.printList2file(file2print, list_fna)
	return (sigfile[0], siglist[0])

	



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez									  ##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain	  ##
##########################################################
'''
Calls FASTQC analysis and parses results generated.
'''
## useful imports
import time
import io
import os
import re
import sys
import fastqcparser
from sys import argv
from io import open
import pandas as pd

import matplotlib
matplotlib.use('agg')
import matplotlib.pyplot as plt
from pandas.plotting import table
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config

############
def	help_options():
	print ("\nUSAGE:\npython %s folder file1 [file2] name fastqc threads\n"  %os.path.realpath(__file__))

############
def call_fastqc(path, file1, file2, sample, fastqc_bin, threads):	
	## call system for fastqc sample given
	#name = functions.create_subfolder(sample, path)
	logFile = path + '/' + sample + '.log'
	
	if os.path.isfile(logFile):
		return ('OK')
	
	if not file2:
		cmd_fastqc = '%s --extract -t %s -o %s %s > %s 2> %s' %(fastqc_bin, threads, path, file1, logFile, logFile)
	else:
        ##print ("+ Calling fastqc for samples...")	
	       cmd_fastqc = '%s --extract -t %s -o %s %s %s > %s 2> %s' %(fastqc_bin, threads, path, file1, file2, logFile, logFile)
	
    ## send command	
	return (functions.system_call( cmd_fastqc ))
		
	
############
def run_module_fastqc(path, files, sample, threads):	
	## Arguments provided via ARGVs
	fastqc_bin = set_config.get_exe('fastqc')

	if (len(files) == 1):
		codeReturn = call_fastqc(path, files[0], "", sample, fastqc_bin, threads)
	elif (len(files) == 2):
		codeReturn = call_fastqc(path, files[0], files[1], sample, fastqc_bin, threads)
	else:
		print("ERROR: Some error ocurred while executing fastqc. Check samples provided")
	
	if codeReturn == 'FAIL':
		exit()
	path_to_sample = path + '/' + sample
	return path_to_sample



############
def main():
	## this code runs when call as a single script

  	## control if options provided or help
	if len(sys.argv) > 1:
		print ("")
	else:
		help_options()
		exit()		

	path = os.path.abspath(argv[1])
	file1 = os.path.abspath(argv[2])
	file2 = os.path.abspath(argv[3])
	sample = argv[4]
	fastqc_bin = argv[5]
	threads = argv[5]

	## check if paired end
	if not (file2):
		print ('+ No implementation yet for single-end. Sorry.')
		exit()

	##
	path_to_sample = call_fastqc(path, file1, file2, sample, fastqc_bin, threads)
		
############
'''******************************************'''
if __name__== "__main__":
	main()










#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Prepares the database information for further analysis. Several functions are implemented for:
	
	- Project data provided, updates/populates the database of interest

'''
## useful imports
import time
import io
import os
import re
import sys
import pandas as pd
import shutil
from sys import argv
from io import open
from termcolor import colored
from Bio import SeqIO
import concurrent.futures

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts import species_identification_KMA
from BacterialTyper.scripts import min_hash_caller
from BacterialTyper.scripts import database_generator
from BacterialTyper.scripts import sampleParser

##########################################################################################
def update_database_user_data(database_folder, project_folder, Debug, options):
	"""
	Updates user_data folder within the database folder provided.
	
	It would generate single subfolder for each sample previously analyzed and it would store main information and result files for later interpretation, comparison and/or summarization with new samples analyzed.
	
	:param database_folder:
	:param project_folder:
	:param Debug:
	:param options:
	
	:type database_folder:
	:type project_folder:
	:type Debug:
	:type options:
	
	:returns: Updated database result from :func:`BacterialTyper.scripts.database_generator.update_db_data_file`.
	:rtype: Dataframe
	
	:warnings: Returns **FAIL** if check process failed.
	
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.create_subfolder`
		
		- :func:`BacterialTyper.scripts.functions.get_data`
		
		- :func:`BacterialTyper.scripts.functions.optimize_threads`
		
		- :func:`BacterialTyper.scripts.database_user.get_userData_files`
		
		- :func:`BacterialTyper.scripts.database_user.update_sample`
		
		- :func:`BacterialTyper.scripts.database_generator.getdbs`
		
		- :func:`BacterialTyper.scripts.database_generator.get_database`
		
		- :func:`BacterialTyper.scripts.database_generator.update_db_data_file`

	"""
	
	print ("\n+ Updating information from user data folder: ", project_folder)
	
	## create folder
	own_data = functions.create_subfolder("user_data", database_folder)
	
	## Default missing options
	options.project = True
	options.debug = Debug
	if not options.single_end:
		options.pair = True

	####################################
	## get information
	####################################
	
	## get user data files
	project_data_df = get_userData_files(options, project_folder)
	
	## get user data info
	project_info_df = get_userData_info(options, project_folder)
	
	## merge data
	project_all_data = pd.concat([project_data_df, project_info_df], join='inner', sort=True).drop_duplicates()
	project_all_data.index.name = 'name'

	## debug messages:
	if Debug:
		print("project_data_df")
		print(project_data_df)
		
		print("project_info_df")
		print(project_info_df)

		print("project_all_data")
		print(project_all_data)
	
	functions.print_sepLine("-", 75, False)
	print ('\n+ Get database information')
	db_frame = database_generator.getdbs('user_data', database_folder, 'user_data', Debug)
	user_data_db = database_generator.get_database(db_frame, Debug)
	
	## merge dataframe
	sample_frame = project_all_data.groupby("name")
	
	####################################
	## optimize threads
	####################################
	name_list = project_all_data.index.values.tolist()
	threads_job = functions.optimize_threads(options.threads, len(name_list)) ## threads optimization
	max_workers_int = int(options.threads/threads_job)

	## debug message
	if (Debug):
		print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
		print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
		print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))

	
	functions.print_sepLine("-", 75, False)
	print ('\n+ Updating information using %s threads and %s parallel jobs' %(options.threads, max_workers_int))

	####################################
	## loop through frame using multiple threads
	####################################
	with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor:
		## send for each	
		commandsSent = { executor.submit(update_sample, name, cluster, own_data, user_data_db, Debug): name for name, cluster in sample_frame }
		for cmd2 in concurrent.futures.as_completed(commandsSent):
			details = commandsSent[cmd2]
			try:
				data = cmd2.result()
			except Exception as exc:
				print ('***ERROR:')
				print (cmd2)
				print('%r generated an exception: %s' % (details, exc))
	
	functions.print_sepLine("+", 75, False)
	print ("\n+ Retrieve information ...")

	####################################
	###### populate dataframe
	####################################
	for name, cluster in sample_frame:
		###### dump to file
		info_file = own_data + '/' + name + '/info.txt'
		if os.path.exists(info_file):
			dataGot = functions.get_data(info_file, ',', 'index_col=0')
			dataGot = dataGot.set_index('ID')
		
			if (options.debug):
				print (colored("**DEBUG: dataGot dataframe **", 'yellow'))
				print (dataGot)
	
			user_data_db = pd.concat([user_data_db, dataGot], join='outer', sort=True).drop_duplicates()
			## concatenating by outer we get all available entries

	if (options.debug):
		print (colored("**DEBUG: user_data_db dataframe **", 'yellow'))
		print (user_data_db)

	functions.print_sepLine("+", 75, False)	

	####################################
	## update db		
	####################################
	database_csv = own_data + '/user_database.csv'
	
	dataUpdated = database_generator.update_db_data_file(user_data_db, database_csv)
	print ("+ Database has been generated: \n", database_csv)
	return (dataUpdated)


##########################################################################################
def get_userData_files(options, project_folder):
	## get information regarding files

	## get assembly files
	pd_samples_assembly = sampleParser.get_files(options, project_folder, "assembly", ["fna"])
	pd_samples_assembly = pd_samples_assembly.set_index('name')

	## get annotation files
	pd_samples_annot = sampleParser.get_files(options, project_folder, "annot", ['gbf', 'faa', 'gff'])
	pd_samples_annot = pd_samples_annot.set_index('name')

	## get trimmed ngs files
	pd_samples_reads = sampleParser.get_files(options, project_folder, "trim", ['_trim'])
	pd_samples_reads = pd_samples_reads.set_index('name')

	## debug message
	if (options.debug):
		print (colored("**DEBUG: pd_samples_reads **", 'yellow'))
		print (pd_samples_reads)
		print (colored("**DEBUG: pd_samples_assembly **", 'yellow'))
		print (pd_samples_assembly)
		print (colored("**DEBUG: pd_samples_annot **", 'yellow'))
		print (pd_samples_annot)
	
	## merge
	df = pd.concat([pd_samples_reads, pd_samples_annot, pd_samples_assembly], sort=True, join='inner').drop_duplicates()
	## joining by inner we only get common columns among all

	## debug message
	if (options.debug):
		print (colored("**DEBUG: pd_concat **", 'yellow'))
		print (df)
	
	## set new column with name of samples
	df = df.reset_index()

	## debug message
	if (options.debug):
		print (colored("**DEBUG: pd_concat reset_index**", 'yellow'))
		print (df)
	##
	return(df)
	
##########################################################################################
def get_userData_info(options, project_folder):
	## get information regarding: 
		## genus, species (ident module)
		## card & VFDB (profile module)
		## additional information: MGE, etc

	## get profile information
	pd_samples_profile = sampleParser.get_files(options, project_folder, "profile", ["csv"])
	if not pd_samples_profile.empty:
		pd_samples_profile = pd_samples_profile.set_index('name')

	## get identification information
	pd_samples_ident = sampleParser.get_files(options, project_folder, "ident", ["csv"])
	if not pd_samples_ident.empty:
		pd_samples_ident = pd_samples_ident.set_index('name')
	
	## get mash information
	pd_samples_mash = sampleParser.get_files(options, project_folder, "mash", ["sig"])
	if not pd_samples_mash.empty:
		pd_samples_mash = pd_samples_mash.set_index('name')

	## add other if necessary

	## debug message	
	if (options.debug):
		print (colored("**DEBUG: pd_samples_profile **", 'yellow'))
		print (pd_samples_profile)
		print (colored("**DEBUG: pd_samples_ident **", 'yellow'))
		print (pd_samples_ident)
		print (colored("**DEBUG: pd_samples_mash **", 'yellow'))
		print (pd_samples_mash)
	
	## merge
	df = pd.concat([pd_samples_profile, pd_samples_ident, pd_samples_mash], join='inner', sort=True).drop_duplicates()
	## joining by inner we only get common columns among all

	## debug message
	if (options.debug):
		print (colored("**DEBUG: pd_concat **", 'yellow'))
		print (df)
	
	## set new column with name of samples
	df = df.reset_index()

	## debug message
	if (options.debug):
		print (colored("**DEBUG: pd_concat reset_index**", 'yellow'))
		print (df)
	##
	return(df)

############################################
def update_sample(name, cluster, own_data, user_data_db, Debug):

	## debug message	
	if (Debug):
		print (colored("**DEBUG: sample_frame groupby: name & cluster **", 'yellow'))
		print (name)			
		print (cluster)
	
	print ('\t+ Sending command for sample: ', name)

	############################################
	#### check information for this sample
	############################################

	## generate sample
	dir_sample = functions.create_subfolder(name, own_data)
	
	if name in user_data_db.index:
		print (colored("\t\t+ Data available in database for sample: %s. Checking integrity..." %name, 'yellow'))
		#functions.print_sepLine("+", 75, False)

	## data to generate
	data2dump = pd.DataFrame(columns=('ID','folder','genus','species','name','genome', 'GFF','proteins', 'signature', 'profile', 'ident', 'reads'))
	## iterate over files with different tags: reads, annot, assembly, profile, ident

	##########
	## assembly
	##########
	assembly_dir = functions.create_subfolder('assembly', dir_sample)
	assembly_file = cluster.loc[cluster['tag'] == 'assembly']['sample'].to_list()
	if assembly_file:
		assembly_file_name = os.path.basename(assembly_file[0])
		genome = assembly_dir + '/' + assembly_file_name
		if not os.path.exists(genome):
			shutil.copy(assembly_file[0], assembly_dir)
	else:
		genome = ""

	##########
	## annot
	##########
	annot_dir = functions.create_subfolder('annot', dir_sample)
	annot_files = cluster.loc[cluster['tag'] == 'annot']['sample'].to_list()
	prof = ""
	gff = ""
	if annot_files:
		for f in annot_files:
			file_name = os.path.basename(f)
			if f.endswith('faa'):
				prot = annot_dir + '/' + file_name
				if os.path.exists(prot):
					continue
			elif f.endswith('gff'):
				gff = annot_dir + '/' + file_name
				if os.path.exists(gff):
					continue
			shutil.copy(f, annot_dir)
	else:
		gff = ""
		prot = ""

	##########
	## trimm
	##########
	trimm_dir = functions.create_subfolder('trimm', dir_sample)
	reads_files = cluster.loc[cluster['tag'] == 'reads']['sample'].to_list()
	reads = []
	if reads_files:
		for f in reads_files:
			file_name = os.path.basename(f)
			reads_name = trimm_dir + '/' + file_name
			reads.append(reads_name)
			if not os.path.exists(reads_name):
				shutil.copy(f, trimm_dir)

	##########
	## ident
	##########
	ident_dir = functions.create_subfolder('ident', dir_sample)
	ident_file = cluster.loc[cluster['tag'] == 'ident']['sample'].to_list()
	if ident_file:
		file_name = os.path.basename(ident_file[0])
		ident_file_name = ident_dir + '/' + file_name
		if not os.path.exists(ident_file_name):
			shutil.copy(ident_file[0], ident_dir)
	else:
		ident_file_name = ""
	
	##########
	## profile
	##########
	profile_dir = functions.create_subfolder('profile', dir_sample)
	profile_files = cluster.loc[cluster['tag'] == 'profile']['sample'].to_list()
	profile_file = []
	if profile_files:
		for f in profile_files:
			file_name = os.path.basename(f)
			profile_file_name = profile_dir + '/' + file_name
			profile_file.append(profile_file_name)
			if not os.path.exists(profile_file_name):
				shutil.copy(f, profile_dir)
	
	##########
	## mash profile
	##########
	mash_dir = functions.create_subfolder('mash', dir_sample)
	mash_file = cluster.loc[cluster['tag'] == 'mash']['sample'].to_list()
	if mash_file:
		file_name = os.path.basename(mash_file[0])
		sig_file = mash_dir + '/' + file_name
		if not os.path.exists(sig_file):
			shutil.copy(mash_file[0], mash_dir)
	else:
		sig_file = ""
	
	############################################
	### Dump information

	## TODO: Add species and genus information when parsed from ident csv file
	#####
	data2dump.loc[len(data2dump)] = (name, dir_sample, 'genus', 'species', name, genome, gff, prot, sig_file, '::'.join(sorted(profile_file)), ident_file_name, '::'.join(sorted(reads)) )
	#data2dump = data2dump.set_index('ID')
	
	###### dump to file
	info_file = dir_sample + '/info.txt'
	data2dump.to_csv(info_file)

	###### dump file information to file
	info_file2 = dir_sample + '/info_files.txt'
	cluster.to_csv(info_file2)

	###### timestamp
	filename_stamp = dir_sample + '/.success'
	stamp =	functions.print_time_stamp(filename_stamp)

	return()




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez					##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain	##
##########################################################
"""
Shared functions used along ``BacterialTyper`` pipeline.
With different purposes:

	- Print time stamps

	- Create system calls

	- Manage/list/arrange files/folders

	- Aesthetics

	- Manage fasta files

	- Other miscellaneous functions
"""
## useful imports
import time
import io
import os
import re
import subprocess
import sys
import wget
import json
from datetime import datetime
from Bio import SeqIO
from termcolor import colored
from filehash import FileHash
import pandas as pd
import patoolib ## to extract
import shutil

## import my modules
from BacterialTyper.config import set_config
from BacterialTyper.other_tools import tools

########################################################################
######## 					TIME								######## 					
########################################################################

###############   
def print_time ():
	"""Prints time stamp in human readable format: month/day/year, hour:minute:seconds."""
	now = datetime.now()
	date_time = now.strftime("%m/%d/%Y, %H:%M:%S")
	print ('\t' + date_time + '\n')

###############   
def gettime (start_time):
	"""Obtains time stamp in human readable format: hour:minute:seconds from a time.time() format timestamp."""
	total_sec = time.time() - start_time
	m, s = divmod(int(total_sec), 60)
	h, m = divmod(m, 60)
	return h, m, s

###############	
def timestamp (start_time_partial):
	"""Prints a stamp of the time spent for a process in human readable format: hour:minute:seconds.
	Returns time in format time.time().
	"""
	h,m,s = gettime(start_time_partial)
	print_sepLine("-", 25, False)
	print ('(Time spent: %i h %i min %i s)' %(int(h), int(m), int(s)))
	print_sepLine("-", 25, False)
	return time.time()

###############	
def print_time_stamp (out):
	"""Prints out timestamp in a file provided. Format: time.time()"""
	timefile = open(out, 'w')    
	string2write = str(time.time())
	timefile.write(string2write)
	return()

###############	
def read_time_stamp (out):
	"""Reads timestamp from a file provided. Format: time.time()"""
	st_hd = open(out, 'r')
	st = st_hd.read()
	st_hd.close()
	stamp = datetime.fromtimestamp(float(st)).strftime('%Y-%m-%d %H:%M:%S')
	return(stamp)

###############	
def get_diff_time(stamp):
	"""Obtains the time spent for a process in days given a stamp in time.time() format.
	Returns days passed since.
	"""

	time_today = time.time()
	elapsed = time_today - float(time_today)
	days_passed = int((elapsed/3600)/24)
	return(days_passed)

###############    
def create_human_timestamp():
	"""Generates human timestamp for the date of day in format (yearmonthday): e.g. 20191011"""
	now = datetime.now()
	timeprint = now.strftime("%Y%m%d")
	return timeprint
###############

############################################################################
######## 					FILES/FOLDERS							######## 					
############################################################################

###############
def is_non_zero_file(fpath):  
	# https://stackoverflow.com/a/15924160
	"""Returns TRUE/FALSE if file exists and non zero"""
	return os.path.isfile(fpath) and os.path.getsize(fpath) > 0

###############
def outdir_project(outdir, project_mode, pd_samples, mode):
	"""
	"""
	# Group dataframe by sample name
	sample_frame = pd_samples.groupby(["new_name"])

	dict_outdir = {}	
	for name, cluster in sample_frame:
		if (project_mode):
			#print ("Create subdir for every sample: ", mode)
			sample_dir = create_subfolder('data', outdir)		

			## create sample
			sample_name_dir = create_subfolder(name, sample_dir)		

			## create subdir sub sample
			mode_name_dir = create_subfolder(mode, sample_name_dir)		
			dict_outdir[name] = mode_name_dir

		else:
			#print ("All samples share same folder")
			sample_name_dir = create_subfolder(name, outdir)		
			dict_outdir[name] = sample_name_dir

	return (dict_outdir)

###############
def outdir_subproject(outdir, pd_samples, mode):
	## we assume we want to create within a project dir a subdir
	# Group dataframe by sample name
	sample_frame = pd_samples.groupby(["name"])
	dict_outdir = {}	
	for name, cluster in sample_frame:
		mode_name_dir = create_subfolder(mode,os.path.join(outdir, name))		
		dict_outdir[name] = mode_name_dir

	return (dict_outdir)

###############
def create_subfolder (name, path):
	"""Create a subfolder named 'name' in directory 'path'. Returns path created."""
	## create subfolder  ##	
	subfolder_path = path + "/" + name
	access_rights = 0o755

	# define the access rights
	try:
		os.mkdir(subfolder_path, access_rights)
	except OSError:  
	   	#print ("\tDirectory %s already exists" % subfolder_path)
		return subfolder_path
	else:  
		print (colored("Successfully created the directory %s " % subfolder_path, 'yellow'))

	return subfolder_path

###############  
def create_folder (path):
	"""Create a folder directory 'path'. Returns path created."""

	## create subfolder  ##	
	access_rights = 0o755

	# define the access rights
	try:
		os.mkdir(path, access_rights)
	except OSError:  
		#print ("\tDirectory %s already exists" %path)
		return path
	else:  
		print (colored("Successfully created the directory %s " %path, 'yellow'))

	return path

############### 
def get_symbolic_link (sample_list, directory):
	"""Creates symbolic links, using system call, for list of files given in directory provided"""
	for samplex in sample_list:
		cmd = 'ln -s %s %s' %(samplex, directory)
		system_call(cmd, returned=False)

	files2return = os.listdir(directory)
	return files2return

############### 
def get_symbolic_link_file (file2link, newfile):
	"""Creates symbolic link for a file into a new name file"""
	cmd = 'ln -s %s %s' %(file2link, newfile)
	system_call(cmd, returned=False)

#################
def get_fullpath_list(dir_given):
	"""Retrieve full absolute path for the files within a directory specified.

	:param dir_given: Directory to retrieve files
	:type dir_given: string

	:returns: List of absolute path files.
	"""
	return_path = []
	for root, dirs, files in os.walk(dir_given):
		for f in files:
			return_path.append(os.path.join(root,f))
	return return_path

#################
def printList2file(fileGiven, listGiven):
	"""Prints list given in the output file provided. One item per row."""
	file_hd = open(fileGiven, 'w')
	file_hd.write("\n".join(listGiven))
	file_hd.close()  

#################
def readList_fromFile(fileGiven):
	"""Reads list from the input file provided. One item per row."""
	# open file and read content into list
	lineList = [line.rstrip('\n') for line in open(fileGiven)]
	return (lineList)

######
def retrieve_matching_files(folder, string):
	"""Lists folder path provided and given a string to search, returns all files ending with the given string"""
	my_all_list = get_fullpath_list(folder)
	matching = [s for s in my_all_list if s.endswith(string)]
	return (matching)

#####
def file2dictionary(file2read, split_char):
	"""Read file and generate a dictionary"""
	d = {}
	with open(file2read) as f:
		for line in f:
			line = line.rstrip('\n')
			(key, val) = line.split(split_char)
			d[key] = val

	return(d)

def file2dataframe(file2read, names):
	## TODO: check if duplicated with get_data function
	"""Read csv file into pandas dataframe"""
	d = pd.read_csv(file2read, comment="#", names=names)
	return(d)


def chmod_rights(file, access_rights):
	"""Changes access permission of a file:
	
	Read additional information in:
	https://docs.python.org/3/library/os.html#os.chmod
	https://www.geeksforgeeks.org/python-os-chmod-method/
	"""
	access = os.chmod(file, access_rights, follow_symlinks=True)


########################################################################
######## 					system call							######## 					
########################################################################

###############
def system_call(cmd, returned=False, message=True):
	"""Generates system call using subprocess.check_output"""
	## call system
	## send command
	if (message):
		print (colored("[** System: %s **]" % cmd, 'magenta'))

	try:
		out = subprocess.check_output(cmd, shell = True)
		if (returned):
			return (out)
		return ('OK')
	except subprocess.CalledProcessError as err:
		if (returned):
			return (err.output)
		if (message):
			print (colored("** ERROR **", 'red'))
			print (colored(err.output, 'red'))
			print (colored("** ERROR **", 'red'))
		
		return ('FAIL')

###############
def wget_download(url, path):
	"""
	Downloads file in the given path.
	
	It uses python module wget to download given ``url`` in the ``path`` provided.
	
	:param url: File to download from http/ftp site.
	:param path: Absolute path to the folder where to save the downloaded file.
	:type url: string
	:type path: string  
	
	:returns: Print messages and generates download
	"""
	
	print ('\t+ Downloading: ', url)
	wget.download(url, path)
	print ('\n')

###############
def check_md5sum(string, File):
	md5hasher = FileHash('md5')
	md5_file = md5hasher.hash_file(File)
	
	if (md5_file == string):
		#print (md5_file + '==' + string)
		return (True)
	else:
		return (False)

###############
def extract(fileGiven, out, remove=True):
	"""
	Extracts archived file
	
	This function extracts the file given in the ``out`` path provided.
	It uses the ``patoolib`` that is able to identify the type of file 
	and compression algorithm to use in each case.
	
	It also removes the compressed file using ``os`` module.
	
	:param fileGiven: Archived file to extract.
	:param out: Output name and absolute path for the extracted archived.
	:param remove: True/False for removing compressed file extracted
	
	:type fileGiven: string
	:type out: string
	:type remove: boolean
	
	"""
	## extract using patoolib
	patoolib.extract_archive(fileGiven, outdir=out, verbosity=0)
	
	if (remove):
		## remove compress file
		print ("Remove compress file...")
		os.remove(fileGiven)
		print ("\n")
	

####################################################################
######## 					BLAST							######## 					
####################################################################

###############
def makeblastdb(DBname, fasta):
	## generate blastdb for genome
	makeblastDBexe = set_config.get_exe('makeblastdb')
	
	if (os.path.isfile(DBname + '.nhr')):
		print ("+ BLAST database is already generated...")
	else:
		cmd_makeblast = "%s -in %s -input_type fasta -dbtype %s -out %s" %(makeblastDBexe, fasta, 'nucl', DBname)
		code = system_call(cmd_makeblast)

		if (code == 'FAIL'):
			print (colored('****ERROR: Some error happened during the makeblastDB command', 'red'))
			print (cmd_makeblast)
			exit()
	

###############	
def blastn(outFile, DBname, fasta, threads):
	# blastn plasmids vs contigs
	blastnexe = set_config.get_exe('blastn')
	cmd_blastn = "%s -db %s -query %s -out %s -evalue 1e-20 -outfmt \'6 std qlen slen\' -num_threads %s" %(blastnexe, DBname, fasta, outFile, threads )
	codeBlastn = system_call(cmd_blastn)
	
	if (codeBlastn == 'FAIL'):
		print (colored('****ERROR: Some error happened during the blastn command', 'red'))
		print (cmd_blastn)
		exit()
	
########################################################################
######## 					FASTA files							######## 					
########################################################################

###############
def concat_fasta(dirFasta, Fasta):
	print ("+ Concatenating all information into one file...")	
	cmd = 'cat ' + dirFasta + '/*fna > ' + Fasta
	return(system_call(cmd))

###############
def subset_fasta(ident, fasta, out):
	output_FASTA = open(out, 'w')	
	for record in SeqIO.parse(fasta, "fasta"):
		all_id = record.description
		species_search = re.search(r"%s" % ident, all_id)
		if species_search:
			head = ">" + all_id + "\n"
			output_FASTA.write(head)
			output_FASTA.write(str(record.seq))
			output_FASTA.write("\n")
	
	output_FASTA.close()

###############
def rename_fasta_seqs(fasta_file, name, new_fasta):
	"""Rename fasta sequences provided in file :file:`fasta_file` using id :file:`name`. Save results in file :file:`new_fasta` provided.
	
	Check for id character lenght do not exceed 37 characters as it might be a limitiation in further annotation and subsequent analysis. Read Prokka_ issue for further details: https://github.com/tseemann/prokka/issues/337.
	
	:param fasta_file: Absolute path to fasta file.
	:type fasta_file: string
	:param name: String to add every fasta sequence header.
	:type name: string
	:param new_fasta: Name for the new fasta file (Absolute path).
	:type new_fasta: string
	:return: Path to tabular delimited file containing conversion from all to new id for each sequence.
	:warnings: Returns FAIL if name is >37 characters.
	
	.. include:: ../../links.inc	 	
	"""

	output_FASTA = open(new_fasta, 'w')
	id_conversion = open(new_fasta + "_conversionID.txt", 'w')
	
	if len(name) > 37:
		print (colored("** ERROR **", 'red'))
		print (colored("BacterialTyper.functions.rename_fasta_seqs():: name id is > 37 characters.", 'red'))
		print (colored("** ERROR **", 'red'))
		return ('FAIL')
	
	counter_seqs = 0
	for record in SeqIO.parse(fasta_file, "fasta"):
		old_id = record.description
		counter_seqs += 1
		new_id = name + "_" + str(counter_seqs)
		head = ">" + new_id + "\n"
		output_FASTA.write(head)
		output_FASTA.write(str(record.seq))
		output_FASTA.write("\n")
		
		id_conversion.write(old_id + "\t" + new_id)
		id_conversion.write("\n")
	
	output_FASTA.close()
	id_conversion.close()
		
	return (new_fasta + "_conversionID.txt")
	

############################################################################
######## 					AESTHETICS								######## 					
############################################################################
def pipeline_header():
	"""
	Prints a common header for the pipeline including name, author, copyright and year.	    
	"""
	print ("\n")
	print_sepLine("#", 70, False)
	print('#', '{: ^66}'.format("BacterialTyper pipeline"), '#')
	print('#', '{: ^66}'.format("Jose F. Sanchez, Cristina Prat & Lauro Sumoy"), '#')
	print('#', '{: ^66}'.format("Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain"), '#')
	print_sepLine("#", 70, False)

###############
def print_sepLine(char, num, color):
	string = char * num
	if (color):
		print (colored(string, color))
	else:
		print (string)

###############
def boxymcboxface(message):
	## this function is from ARIBA (https://github.com/sanger-pathogens/ariba)
	## give credit to them appropiately
   	#print('-' * 79)
	print ('\n')
	print('|', '=' * 50, '|', sep='')
	print('|', '{: ^48}'.format(message), '|')
	print('|', '=' * 50, '|', sep='')
	print ('\n')
    #print('-' * 79)

###############
def progbar(curr, total, full_progbar):
	frac = (curr/total)
	filled_progbar = round(frac*full_progbar)
	print ('\r', '#'*filled_progbar + '-'*(full_progbar-filled_progbar), '[{:>7.2%}]'.format(frac), end='')
	sys.stdout.flush()


########################################################################
######## 					Miscellaneous						######## 					
########################################################################

###############
def get_data(ID_file, SEP, options):	
	if options == 'index_col=0':
		data = pd.read_csv(ID_file, sep=SEP, index_col=0)
		return(data)
	else:
		data = pd.read_csv(ID_file, sep=SEP)
		return(data)
	## fix for another example if any
	
###############
def get_number_lines(input_file):	
	with open(input_file) as foo:
		lines = len(foo.readlines())
	foo.close()
	return (lines)

###############
def get_info_file(input_file):	
	with open(input_file) as f:
	    lines = f.read().splitlines() 
	f.close()
	return (lines)
  	    
#################
def optimize_threads(total, samples):
	cpu = int(int(total)/int(samples))
	
	if (cpu==0): ## 5 availabe cpus for 10 samples == 0 cpu
		cpu = 1
	
	return(cpu)

#################
def parse_sublist(lst, ind): 
	## extract elemnts of the list
	## Original Code: https://www.geeksforgeeks.org/python-get-first-element-of-each-sublist/
	return [item[ind] for item in lst]

##################
def decode(x):
	"""
	Python decode string method

	It converts bytes to string.

	:param x: String of text to decode
	:type x: string
	:returns: Text decoded

	.. attention:: Be aware of Copyright

		The code implemented here was retrieved and modified from ARIBA (https://github.com/sanger-pathogens/ariba)

		Give them credit accordingly.
	"""
	try:
		s = x.decode()
	except:
		return x

	return s

###################
def collect_info_run(out_folder, module, userinput, runInfo):
	"""
	Prints information of the job run in json format
	
	:param out_folder: Absolute path to folder to dump results
	:param module: Module or script name executed
	:param userinput: Contains information about the input parameters
	:param runInfo: Contains configuration detail information to dump
	
	:type out_folder: string 
	:type module: string 
	:type userinput: dict
	:type runInfo: dict
	
	:returns: Prints information in json format in the output folder provided.
	
	Exampple of infromation to include:
	
	userinput = {"filename":infiles, 
				"database":path_to_database,
				"file_format":file_format}
				
	runInfo = { "module":module, 
				"analysis":example,
				"date":date, 
				"time":time}
	
	Original idea extracted from https://bitbucket.org/genomicepidemiology/kmerfinder/src/master/kmerfinder.py
	
	"""
	
	# Collect all data of run-analysis
	data = {service:{}}
	
	data[service]["user_input"] = userinput
	data[service]["run_info"] = runInfo
	
	# Save json output
	result_json_file = out_folder + "/" + module + ".json" 
	with open(result_json_file, "w") as outfile:  
	   json.dump(data, outfile)
	
def print_all_pandaDF(pd_df):
	pd.set_option('display.max_colwidth', None)
	pd.set_option('display.max_columns', None)
	print (pd_df)



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Prints an index of citation for the different packages and other softwares employed here.
"""

## import my modules
from BacterialTyper.config import set_config
from BacterialTyper.scripts import functions
from termcolor import colored
import os, sys
import pandas as pd

#########
def software_citation():
	soft_citation ={
		#'CARD':['The Comprehensive Antibiotic Resistance Database', 'McArthur et al 2013', 'PMID: 23650175', 'https://card.mcmaster.ca/'],
		'ARIBA':['-', '-', '-', '-'],
		'AUGUSTUS':['-', '-', '-', '-'],
		'Bowtie2':['-', '-', '-', '-'],
		'BLAST':['-', '-', '-', '-'],
		'cdhit':['-', '-', '-', '-'],
		'FASTQC':['-', '-', '-', '-'],
		'KMA':['-', '-', '-', '-'],
		'PROKKA':['-', '-', '-', '-'],
		'SPADES':['-', '-', '-', '-'],
		'TRIMMOMATIC':['-', '-', '-', '-']		
	}
	return(soft_citation)

#########
def ariba_citation():
	ARIBA_DB_citation ={
		'CARD':['The Comprehensive Antibiotic Resistance Database', 'McArthur et al 2013', 'PMID: 23650175', 'https://card.mcmaster.ca/'],
		'MEGARes':['MEGARes: an antimicrobial database for high throughput sequencing', 'Lakin et al 2016', 'PMID: 27899569', 'http://megares.meglab.org/'],
		'PlasmidFinder':['PlasmidFinder and pMLST: in silico detection and typing of plasmids', 'Carattoli et al 2014', 'PMID: 24777092', '-'],
		'ResFinder':['Identification of acquired antimicrobial resistance genes', 'Zankari et al 2012', 'PMID: 22782487', '-'],
		'ARG-ANNOT':['ARG-ANNOT: a new bioinformatic tool to discover antibiotic resistance genes in bacterial genomes', 'Gupta et al 2014', 'PMID: 24145532', '-'],
		'srst2':['SRST2: Rapid genomic surveillance for public health and hospital microbiology labs', 'Inouye et al 2014', 'PMID: 25422674', 'https://github.com/katholt/srst2'],
		'VFDB':['VFDB 2019: a comparative pathogenomic platform with an interactive web interface.', 'Liu et al 2019', 'PMID: 30395255', 'http://www.mgc.ac.cn/VFs/main.htm'],
		'VirulenceFinder':['Real-time whole-genome sequencing for routine typing surveillance and outbreak detection of verotoxigenic Escherichia coli', 'Joensen al 2014', 'PMID: 24574290', '-']
	}
	return(ARIBA_DB_citation)

#########
def print_all():
	print ("")
	functions.print_sepLine("+", 50, 'yellow')
	print ("\tSOFTWARE")
	functions.print_sepLine("+", 50, 'yellow')
	print ("Third party softwares included or employed during the pipeline workflow.")
	print ("")
	df_software_citation = pd.DataFrame.from_dict(software_citation(), orient='index', columns=('Article Title', 'Authors', 'PUBMED ID', 'Website'))	
	df_software_citation.index.names = ['Software']
	pd.set_option('display.max_colwidth', None)
	pd.set_option('display.max_columns', None)
	print (df_software_citation)
	print ("")
	
	functions.print_sepLine("+", 50, 'yellow')
	print ("\tDATABASES")
	functions.print_sepLine("+", 50, 'yellow')
	print ("")
	print ("Please cite according to your selection.")
	print ("")
	
	functions.print_sepLine("+", 50, False)
	print ("\tARIBA databases")
	functions.print_sepLine("*", 50, False)	
	df_ARIBA_DB_citation = pd.DataFrame.from_dict(ariba_citation(), orient='index', columns=('Article Title', 'Authors', 'PUBMED ID', 'Website'))	
	df_ARIBA_DB_citation.index.names = ['Databases']
	print (df_ARIBA_DB_citation)
	print ("\n")
	
	functions.print_sepLine("*", 50, False)
	print ("\tKMA software & databases")
	functions.print_sepLine("*", 50, False)

	print ()	
	print ()	

	functions.print_sepLine("*", 50, False)
	print ("\tBUSCO software & dataset")
	functions.print_sepLine("*", 50, False)
	print ("BUSCO applications from quality assessments to gene prediction and phylogenomics.")
	print ("Robert M. Waterhouse, Mathieu Seppey, Felipe A. Simão, Mose Manni, Panagiotis ")
	print ("Ioannidis, Guennadi Klioutchnikov, Evgenia V. Kriventseva, and Evgeny M. Zdobnov")
	print ("Mol Biol Evol, published online Dec 6, 2017, doi: 10.1093/molbev/msx319")
	print ()	
	print ("BUSCO: assessing genome assembly and annotation completeness with single-copy orthologs.")
	print ("Felipe A. Simão, Robert M. Waterhouse, Panagiotis Ioannidis, Evgenia ")
	print ("V. Kriventseva, and Evgeny M. Zdobnov")
	print ("Bioinformatics, published online June 9, 2015, doi: 10.1093/bioinformatics/btv351")
	print ()	
	print ("For further details, please visit: https://busco.ezlab.org/ or https://www.orthodb.org/")

	print ()	
	print ()		
	
	
#########
def only_us():
	print ("Citation to be included when properly finished...\n\n")

#########
def run(options):
	functions.pipeline_header()
	functions.boxymcboxface("Citation")
	if (options.option == 'all'):
		print (colored("\n+ BacterialTyper citation:", 'blue'))
		only_us()

		print (colored("\n+ Other softwares employed in the pipeline", 'blue'))
		print_all()
		
	elif (options.option == 'only'):
		print (colored("\n+ BacterialTyper citation:", 'blue'))
		only_us()
	
	return()
	



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Adds metadata information for each sample analyzed and prepares data for later visualization.
"""
## import useful modules
import os
import sys
import re
import time
from termcolor import colored

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts import database_generator
from BacterialTyper.scripts import database_user

##############################################
def run(options):

	## init time
	start_time_total = time.time()

	##################################
	### show help messages if desired	
	##################################
	if (options.help_metadata):
		## help_format option
		help_metadata()
		exit()


#usr/bin/en python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Retrieves files within ``other_tools`` directory and returns path to given script specified
"""
## useful imports
import os
## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import install_dependencies

####################################################################
def perl_scripts(script):
	"""Lists files within ``other_tools/perl`` directory and returns path to given script.
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.get_fullpath_list`
	"""
	perlDir = os.path.dirname(os.path.realpath(__file__)) + '/perl/'
	list_perl = functions.get_fullpath_list(perlDir)
	
	dict_perl = {}
	for f in list_perl:
		name = os.path.splitext(os.path.basename(f))[0]
		if (name == script):
			return (f)		

####################################################################
def R_scripts(script):
	"""Lists files within ``other_tools/R`` directory and returns path to given script
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.get_fullpath_list`
	"""
	RDir = os.path.dirname(os.path.realpath(__file__)) + '/R/'
	list_R = functions.get_fullpath_list(RDir)
	
	dict_R = {}
	for f in list_R:
		name = os.path.splitext(os.path.basename(f))[0]
		if (name == script):
			return (f)
		


		
	


#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Call KMA_ software to find the best match in reads file or fasta file in a (kmer) database produced using the KMA program 

.. include:: ../../links.inc 
"""
## useful imports
import time
import io
import os
import re
import sys
import pandas as pd
from sys import argv
from io import open
from termcolor import colored
import shutil

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts import ariba_caller

########################
#### INFORMATION	####
########################

##################################################
def help_kma_database():
	"""
	KMA_ software description message.
	"""
	## [TODO]
	print (colored("\n\n***** TODO: Generate this help message *****\n\n", 'red'))

####################
#### DATABASE	####
####################

##################################################
def download_kma_database(folder, database, debug):
	"""
	Downloads databases from KMA website.
	
	Using the latest available ftp datasets, this function downloads available datasets using
	function :func:`BacterialTyper.scripts.functions.wget_download`. 
	
	Ftp site: "ftp://ftp.cbs.dtu.dk/public/CGE/databases/KmerFinder/version/latest/"
	
	It also downloads the md5sum for the dataset selected and compares with the 
	
	:param folder: Absolute path to folder that contains database.
	:param database: Possible options: [bacteria, archaea, protozoa, fungi, plasmids, typestrains, viral].
	:param debug: True/false for printing debugging messages.
	
	:type folder: string
	:type database: string
	:type debug: boolean
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.scripts.functions.wget_download`

		- :func:`BacterialTyper.scripts.functions.check_md5sum`

		- :func:`BacterialTyper.scripts.functions.extract`
		
		- :func:`BacterialTyper.scripts.functions.print_time_stamp`
		
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`

		- :func:`BacterialTyper.scripts.species_identification_KMA.check_db_indexed`

	"""

	## ToDo: update with latest version
	ftp_site = "ftp://ftp.cbs.dtu.dk/public/CGE/databases/KmerFinder/version/latest/"
	
	## In v20190107 there was a plasmid database.
	#ftp_site = "ftp://ftp.cbs.dtu.dk/public/CGE/databases/KmerFinder/version/20190107/"

	## ToDo: Set automatic: download config file and look for prefix for each sample and generate a dictionary to code the prefix for each db.
	
	# Database configuration file - Describes the content of the database
	# Each db consist of 5 files with the following extensions: b, comp.b, length.b, seq.b, name
	# Other important files are: .name, .kma.entries.all, .kma.entries.deleted, .kma.entries.added, .md5
	# db_prefix	name	description
	#bacteria.ATG	Bacteria Organisms	Bacteria organisms library prefix=ATG
	#plasmids.T	Bacteria Plasmids	Bacteria plasmids library prefix=T
	#typestrains.ATG	Bacteria Type Strains	Bacteria type strains library prefix=ATG
	#fungi.ATG	Fungi	Fungi library prefix=ATG
	#protozoa.ATG	Protozoa	Protozoa library prefix=ATG
	#archaea.ATG	Archaea	Archaea library prefix=ATG	
	
	## debug message
	if (debug):
		print (colored("Function call: download_kma_database " + folder + ' ' + database + '\n','yellow'))

	## prefix
	if (database == 'plasmids'):
		prefix = '.T'
	elif (database == 'viral'):
		prefix = '.TG'
	else:
		prefix = '.ATG'
		
	index_name = folder + '/' + database + prefix	

	## check if already download
	return_code_down = False
	if os.path.exists(folder):
		return_code_down = check_db_indexed(index_name, folder)
		## debug message
		if (debug):
			print (colored("Folder database is already available:" + folder,'yellow'))
		
	if (return_code_down == False): ## folder does not exists

		## Download data
		print ("\t+ Downloading data now, it may take a while....")

		## debug message
		if (debug):
			print (colored("Download files via function wget_download:",'yellow'))
		
		## connect to url
		url = ftp_site + database + '.tar.gz'
		functions.wget_download(url, folder)

		md5_url = ftp_site + database + '.md5'
		functions.wget_download(md5_url, folder)
		print ("\n\t+ Data succesfully downloaded.....")

		## get files
		files = os.listdir(folder)
		md5_sum = ""
		for f in files:
			if f.endswith('tar.gz'):
				tar_file = folder + '/' + f
			elif f.endswith('md5'):
				md5_sum = folder + '/' + f
		
		## check md5sum
		print ("\t+ Checking for integrity using md5sum")
		
		# get md5 sum from source
		md5_string = ""
		with open(md5_sum, 'r') as myfile:
			line = myfile.read()
		
		line = re.sub(r"\s", ',', line)
		md5_string = line.split(",")[0]
		
		## calculate md5 for file
		result_md5 = functions.check_md5sum(md5_string, tar_file)
		if (result_md5 == True):
		
			## debug message
			if (debug):
				print (colored("result md5sum matches code provided for file " + tar_file,'yellow'))

			# extract
			print ("\t+ Extracting database into destination folder: " + folder)
			functions.extract(tar_file, folder)	

		else:
			print (colored("*** ERROR: Some error occurred during the downloading and file is corrupted ***", 'red'))
			return ("Error")
			
		## database should be unzipped and containing files...
		return_code_extract = check_db_indexed(index_name, folder)
		
		if (return_code_extract):
			print("+ Database (%s) successfully extracted in folder: %s..." %(database, folder))
		else:
			string = "*** ERROR: Some error occurred during the extraction of the database (%s). Please check folder (%s) and downloading and file is corrupted ***" %(database, folder)
			print (colored(string, 'red'))
			return ("Error")
		
		## print timestamp
		filename_stamp = folder + '/.success'
		stamp =	functions.print_time_stamp(filename_stamp)


##################################################
def check_db_indexed(index_name, folder):
	"""
	Check the status of a database
	
	:param index_name: Index name for the database
	:param folder: Absolute path of the folder containing the database.
	
	:type index_name: string
	:type folder: string
	
	:returns: True/False for the index status.
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
		
		- :func:`BacterialTyper.scripts.functions.readList_fromFile`
		
		- :func:`BacterialTyper.scripts.functions.get_number_lines`
		
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`

		- :func:`BacterialTyper.scripts.functions.print_time_stamp`
	
	"""
	
	# Each db consist of 5 files with the following extensions: b, comp.b, length.b, seq.b, name
	my_index_list = [".comp.b", ".index.b", ".length.b", ".name", ".seq.b"]

	print ("\t+ Checking if database has been previously indexed...")
	for sufix in my_index_list:
		##print (sufix)
		
		my_file = index_name + sufix
		if os.path.isfile(my_file):
			print ("\t" + my_file + ' exists...')
		else:
			if (sufix == '.index.b'):
				continue
			else:
				return(False)
	
	## check if previously assembled and succeeded
	filename_stamp = folder + '/.success'
	if os.path.isfile(filename_stamp):
		stamp =	functions.read_time_stamp(filename_stamp)
		print (colored("\tDatabase was generated on: %s" %stamp, 'yellow'))
		## [TODO]: Check if necessary to download again after several months/days
	
	## dump in screen
	names = index_name + '.name'
	count = functions.get_number_lines(names)
	
	print ("\n\t+ Database seems OK and contains several entries (%s):\n" %count)
	if (count > 50):
		print ("\tToo many entries in the database.\n\tCheck file %s for further details." %names)
	else:
		entries = functions.readList_fromFile(names)
		print (*entries, sep='\n')

	return(True)
	
##################################################
def index_database(fileToIndex, kma_bin, index_name, option, folder, type_option):
	"""
	Calls KMA_ software to index fasta files into a database for later KMA identification.
	
	:param fileToIndex: Fasta file to include in the database.
	:param kma_bin: Absolute path to kma executable binary. 
	:param index_name: Name for the database.
	:param option: Option to create or update a database.
	:param folder: Absolute path to folder containing database.
	:param type_option: Option to index the database with batch: batch [Default: Off].
	
	:type fileToIndex: string
	:type kma_bin: string 
	:type index_name: string 
	:type option: string 
	:type folder: string 
	:type type_option: string 	 
	
	:returns: It returns message from :func:`BacterialTyper.scripts.species_identification_KMA.check_db_indexed`.
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.scripts.functions.create_folder`
		
		- :func:`BacterialTyper.scripts.functions.system_call`
		
		- :func:`BacterialTyper.scripts.species_identification_KMA.check_db_indexed`
	"""	
	
	########################################################################################
	## 								KMA_index-1.2.2							
	########################################################################################
	# kma_index creates the databases needed to run KMA, from a list of fasta files given.
	# Options are:		
	#				Desc:									Default:
	#
	#	-i			Input/query file name (STDIN: "--")		None
	#	-o			Output file								Input/template file
	#	-batch		Batch input file
	#	-deCon		File with contamination (STDIN: "--")	None/False
	#	-batchD		Batch decon file
	#	-t_db		Add to existing DB						None/False
	#	-k			Kmersize								16
	#	-k_t		Kmersize for template identification	16
	#	-k_i		Kmersize for indexing					16
	#	-ML			Minimum length of templates	kmersize 	(16)	
	#	-CS			Start Chain size						1 M
	#	-ME			Mega DB									False
	#	-NI			Do not dump *.index.b					False
	#	-Sparse		Make Sparse DB ('-' for no prefix)		None/False
	#	-ht			Homology template						1.0
	#	-hq			Homology query							1.0
	#	-and		Both homolgy thresholds
	#				has to be reached						or
	#	-v			Version
	#	-h			Shows this help message
	#######################################################################################
	
	## check if file exists
	if os.path.isfile(index_name):
		index_file_name = index_name
	else:
		index_file_name = folder + '/' + index_name
		
	logFile = index_file_name + '.log'
	
	## check if folder exists
	functions.create_folder(folder)

	## single file
	if (type_option == 'batch'):
		type_option = '-batch'
	else:
		type_option = '-i'
	
	## new or add to existing db
	if (option == "new"):
		print ("\n+ Generate and index database for kmer alignment search...\n")
		cmd_kma_index = "%s index %s %s -o %s 2> %s" %(kma_bin, type_option, fileToIndex, index_file_name, logFile)
	elif (option == "add"):
		print ("\n+ Updating database with new entries...\n")
		cmd_kma_index = "%s index %s %s -o %s -t_db %s 2> %s" %(kma_bin, type_option, fileToIndex, index_file_name, index_file_name, logFile)

	code = functions.system_call(cmd_kma_index)	
	if code == 'FAIL':
		print (colored("Database generated an error during the index: %s" %index_name, 'red'))
		print (colored("EXIT", 'red'))
		exit()
		
	return_code = check_db_indexed(index_file_name, folder)
	return(return_code)

##################################################
def generate_db(file_abs_paths, name, fold_name, option, type_option, Debug, kma_bin):
	"""Generate a call to create or update index KMA databases for later kmer identification. 

	:param file_abs_paths: List of absolute paths fasta genome files to include in the database.
	:param name: Database name.
	:param fold_name: Directory path to store database generated.
	:param option: Generate a new database (option = 'new') or add to pre-existing database (option = 'add'). If database exists, automatically adds.
	:param type_option: Index genome fasta files one by one (option_type='single') or using a batch file containing multiple entries (option='batch').
	:param kma_bin:	Binary executable for KMA software 
	:param Debug: True/False for debugging messages.
	
	:type file_abs_paths: list
	:type name: string
	:type fold_name: string
	:type option: string
	:type type_option: string 
	:type kma_bin:
	:type Debug: bool
		
	:returns: Absolute path to database generated
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.scripts.functions.readList_fromFile`
		
		- :func:`BacterialTyper.scripts.functions.printList2file`
		
		- :func:`BacterialTyper.scripts.species_identification_KMA.check_db_indexed`

		- :func:`BacterialTyper.scripts.species_identification_KMA.index_database`
		
	"""

	print ('+ Updating the KMA database: ', name)			

	## check
	if len(file_abs_paths) > 1:
		## read db in fold_name and get index files
		info = fold_name + '/' + name + '.db'

		## 
		lineList = []
		toIndexList = []
		indexedList = []		

		###
		if os.path.exists(info):
			lineList = functions.readList_fromFile(info)
			option = 'add'

		for f in file_abs_paths:
			baseName = os.path.basename(f)
			
			## check if already index
			if baseName in lineList:
				print (colored('\t+ File %s is already available in database %s' %(baseName, name), 'green'))
				indexedList.append(f)
			else:
				toIndexList.append(f)		
		
		if toIndexList:
			## generate batch and call
			info2 = fold_name + '/.batch_entries.txt'
			functions.printList2file(info2, toIndexList)
			status = index_database(info2, kma_bin, name, option, fold_name, type_option)
			final_list = set(lineList + toIndexList + indexedList)
			final_list_name = [os.path.basename(f) for f in final_list]
			functions.printList2file(info, final_list_name)
			count_files = len(toIndexList)
			print ('+ %s samples have been added to the database' %count_files)
		else:
			print ('\n+ No new sequences were added to the database.')
			return (fold_name + '/' + name)			
		
	else:
		file_name = file_abs_paths[0]
		## check if previously indexed
		status = check_db_indexed(file_name, fold_name)
		if (status): #true
			## debug message
			if (Debug):
				print (colored("**DEBUG: Database (%s) is indexed" %file_name + " **", 'yellow'))
			return (file_name)
		else: #false
			## debug message
			if (Debug):
				print (colored("**DEBUG: Database (%s) is not indexed" %file_name + " **", 'yellow'))
			status = index_database(file_name, kma_bin, file_name, option, fold_name, type_option)
	
	## return
	if (status): #true
		return (file_name)
	else:
		return False

def load_db(kma_bin, db2use):
	"""
	This function loads the given database in memory.
	
	:param kma_bin: Absolute path to KMA binary.
	:param db2use: Database to load in memory
	
	:type kma_bin: string
	:type db2use: string 
	
	:returns: System call status. 
	"""
	cmd_load_db = "%s shm -t_db %s -shmLvl 1" %(kma_bin, db2use)
	return_code_load = functions.system_call(cmd_load_db)
	
	return(return_code_load)

def remove_db(kma_bin, db2use):
	"""
	This function removes the given database from memory.
	
	:param kma_bin: Absolute path to KMA binary.
	:param db2use: Database to remove from memory
	
	:type kma_bin: string
	:type db2use: string 
	
	:returns: System call status. 
	"""

	cmd_rm_db = "%s shm -t_db %s -shmLvl 1 -destroy" %(kma_bin, db2use)
	return_code_rm = functions.system_call(cmd_rm_db)
	
	return(return_code_rm)
			

########################
#### IDENTIFICATION	####
########################

##################################################
def kma_ident_call(out_file, files, sample_name, index_name, kma_bin, option, threads):
	"""Create kma system call for kmer identification. 
	
	Paired-end end or single end fastq files accepted. It generates a time stamp if succeeds.

	:param out_file: Absolute path and basename for the output files generated with results.
	:param files: List of absolute paths for fastq files to search againts the database.
	:param sample_name: Directory path to store database generated.
	:param index_name: Database name
	:param kma_bin: Binary executable for KMA software.
	:param option: Additional options to pass to the system call.
	:param threads: Number of CPUs to use. 

	:type out_file: string
	:type files: list
	:type sample_name: string
	:type index_name: string
	:type kma_bin: string
	:type option: string
	:type threads: integer	

	:returns: System call returned finish status.
	
	.. seealso:: This function depends on other ``BacterialTyper`` functions called:
	
		- :func:`BacterialTyper.scripts.functions.system_call`
	
		- :func:`BacterialTyper.scripts.functions.print_time_stamp`

	"""

	###
	out_file_log = out_file + '.log'
	if len(files) == 2:
		cmd_kma_search = "%s -ipe %s %s -o %s -t_db %s -t %s %s 2> %s" %(kma_bin, files[0], files[1], out_file, index_name, threads, option, out_file_log)
	else:
		## TODO: test Single End
		cmd_kma_search = "%s -i %s -o %s -t_db %s -t %s %s 2> %s" %(kma_bin, files[0], out_file, index_name, threads, option, out_file_log)

	code = functions.system_call(cmd_kma_search)

	if (code == 'OK'):
		## success stamps
		basename_tag = os.path.basename(out_file)
		folder = os.path.dirname(out_file)
		filename_stamp = folder + '.success_' + basename_tag
		stamp =	functions.print_time_stamp(filename_stamp)
		return('OK')
	else:
		return('FAIL')

##################################################
def parse_kma_results(out_file, cutoff):
	"""Filters KMA results given a cutoff threshold.
	
	:param out_file: Results file originated from kma system call (kma_ident_call)
	:param cutoff: Template coverage cutoff.

	:type out_file: string
	:type cutoff: integer	

	:returns: Filtered dataframe of results.
	"""
	results = pd.read_csv(out_file, sep="\t")
	results_filter = results[results['Template_Coverage'] > cutoff] 
	## filter according to coverage of the template. 
	return (results_filter)

##################################################
def	help_options():
	print ("\nUSAGE: python %s name fasta folder_name sample_name threads files_csv_string\n"  %os.path.realpath(__file__))

##################################################
def main():
	## this code runs when call as a single script

  	## control if options provided or help
	if len(sys.argv) > 1:
		print ("")
	else:
		help_options()
		exit()    	
	
	## arguments
	name = argv[1]
	fasta = os.path.abspath(argv[2])
	folder = os.path.abspath(argv[3])
	sample_name = argv[4]
	threads = argv[5]

	files = []
	for i,e in enumerate(argv):
		if i > 5:
			files.append(os.path.abspath(argv[i]))

	## other
	cutoff=80
	kma_bin = set_config.get_exe("kma")
	out_file = sample_name + ".out_kma-search.txt"

	## check if database is indexed
	if not  check_db_indexed(name, folder):
		index_database(fasta, kma_bin, name, 'new', folder, '')
		print ("\n+ Database indexed")
	

	## search files
	kma_ident_call(out_file, files, sample_name, folder + '/' + name, kma_bin, '', threads)

	##
	#parse_kma_results(file_out, cutoff)

##################################################
if __name__== "__main__":
	main()

################################################################################################
## 										KMA v1.2.2
################################################################################################	
## read additional information in: https://bitbucket.org/genomicepidemiology/kma/src/master/KMAspecification.pdf
# KMA-1.2.2 mapps raw reads to a template database.
# Options are:		
##				Desc:													Default:	Requirements:
#
#	-o			Output file												None		REQUIRED
#	-t_db		Template DB												None		REQUIRED
#	-i			Input file name(s)										STDIN
#	-ipe		Input paired end file name(s)
#	-int		Input interleaved file name(s)
#	-k			Kmersize												DB defined
#	-e			evalue													0.05
#	-ConClave	ConClave version										1
#	-mem_mode	Use kmers to choose best template, and save memory		False
#	-ex_mode	Searh kmers exhaustively								False
#	-ef			Print additional features								False
#	-vcf		Make vcf file, 2 to apply FT							False/0
#	-deCon		Remove contamination									False
#	-dense		Do not allow insertions in assembly						False
#	-ref_fsa	Consensus sequnce will have "n" instead of gaps			False
#	-matrix		Print assembly matrix									False
#	-a			Print all best mappings									False
#	-mp			Minimum phred score										20
#	-5p			Cut a constant number of nucleotides from the 5 prime.	0
#	-Sparse		Only count kmers										False
#	-Mt1		Map only to "num" template.								0 / False
#	-ID			Minimum ID												1.0%
#	-ss			Sparse sorting (q,c,d)									q
#	-pm			Pairing method (p,u,f)									u
#	-fpm		Fine Pairing method (p,u,f)								u
#	-apm		Sets both pm and fpm									u
#	-shm		Use shared DB made by kma_shm							0 (lvl)
#	-1t1		Skip HMM												False
#	-ck			Count kmers instead of pseudo alignment					False
#	-ca			Make circular alignments								False
#	-boot		Bootstrap sequence										False
#	-bc			Base calls should be significantly overrepresented.		[True]
#	-bc90		Base calls should be both  significantly 
#				overrepresented, and have 90% agreement.				False
#	-bcNano		Call bases at suspicious deletions, made for nanopore.	False
#	-bcd		Minimum depth at base									1
#	-bcg		Maintain insignificant gaps
#	-and		Both mrs and p_value thresholds has 
#				to reached to in order to report a template hit
#	-mq			Minimum mapping quality									0
#	-mrs		Minimum alignment score, normalized to 
#				alignment length										0.50
#	-reward		Score for match											1
#	-penalty	Penalty for mismatch									-2
#	-gapopen	Penalty for gap opening									-3
#	-gapextend	Penalty for gap extension								-1
#	-per		Reward for pairing reads								7
#	-cge		Set CGE penalties and rewards							False
#	-t			Number of threads										1
#	-v			Version
#	-h			Shows this help message
#################################################################################################

## kma index -batch ../database.txt -k 30 -k_t 30 -k_i 30 -o index_db_k30
## kma -ipe WTCHG_370809_205154/WTCHG_370809_205154_trim_R1.fastq WTCHG_370809_205154/WTCHG_370809_205154_trim_R2.fastq -o kma_search3 -t_db index_kma/index_db_k30 -t 3 -apm u -1t1 -k 30
	




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Generates sample quality control at raw reads, assembly or annotation level.
'''
## useful imports
import time
import io
import os
import re
import sys
from io import open
import concurrent.futures
import pandas as pd
import shutil
from termcolor import colored

## import my modules
from BacterialTyper.scripts import fastqc_caller
from BacterialTyper.scripts import sampleParser
from BacterialTyper.scripts import multiQC_report
from BacterialTyper.scripts import BUSCO_caller
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.modules import help_info

################################################
def run_QC(options):

	## init time
	start_time_total = time.time()

	##################################
	### show help messages if desired	
	##################################
	if (options.help_format):
		## help_format option
		sampleParser.help_format()
		exit()
	elif (options.help_BUSCO):
		## information for BUSCO
		BUSCO_caller.print_help_BUSCO()
		exit()
	elif (options.help_project):
		## information for project
		help_info.project_help()
		exit()
	elif (options.help_multiqc):
		## information for Multiqc
		multiQC_report.multiqc_help()
		exit()
		
	## debugging messages
	global Debug
	if (options.debug):
		Debug = True
	else:
		Debug = False	
	
	### set as default paired_end mode
	if (options.single_end):
		options.pair = False
	else:
		options.pair = True
	
	## set main header
	functions.pipeline_header()
	functions.boxymcboxface("Quality check")
	print ("--------- Starting Process ---------")
	functions.print_time()

	## absolute path for in & out
	input_dir = os.path.abspath(options.input)
	outdir=""

	## Project mode as default
	if (options.detached):
		options.project = False
		outdir = os.path.abspath(options.output_folder)
	else:
		options.project = True
		outdir = input_dir		
	
	### option
	if (options.raw_reads):
		## get files
		pd_samples_retrieved = sampleParser.get_files(options, input_dir, "fastq", ("fastq", "fq", "fastq.gz", "fq.gz"))
		fastqc(pd_samples_retrieved, outdir, options, start_time_total, "raw", Debug)
	elif (options.trim_reads):
		## get files
		pd_samples_retrieved = sampleParser.get_files(options, input_dir, "trim", ['_trim'])
		fastqc(pd_samples_retrieved, outdir, options, start_time_total, "trimmed", Debug)
	elif (options.assembly):
		BUSCO_check(input_dir, outdir, options, start_time_total, "genome")
	elif (options.annotation):
		BUSCO_check(input_dir, outdir, options, start_time_total, "proteins")
		
	return()

################################################
def fastqc(pd_samples_retrieved, outdir, options, start_time_total, name_analysis, Debug):
	
	functions.boxymcboxface("FASTQC Quality check for samples")
	
	## debug message
	if (Debug):
		print (colored("\n**DEBUG: pd_samples_retrieve **", 'yellow'))
		print (pd_samples_retrieved)
		print ("\n")

	## generate output folder, if necessary
	print ("\n+ Create output folder(s):")
	
	## if not project, outdir contains the dir to put output
	## in this case, in some other cases might not occur	
	if not options.project:
		functions.create_folder(outdir)
	outdir_dict = functions.outdir_project(outdir, options.project, pd_samples_retrieved, "fastqc_" + name_analysis)
	
	print ("+ Checking quality for each sample retrieved...")
	start_time_partial = start_time_total
	
	# Group dataframe by sample name
	sample_frame = pd_samples_retrieved.groupby(["name"])

	## optimize threads
	name_list = set(pd_samples_retrieved["name"].tolist())
	threads_job = functions.optimize_threads(options.threads, len(name_list)) ## threads optimization
	max_workers_int = int(options.threads/threads_job)

	## debug message
	if (Debug):
		print (colored("**DEBUG: options.threads " +  str(options.threads) + " **", 'yellow'))
		print (colored("**DEBUG: max_workers " +  str(max_workers_int) + " **", 'yellow'))
		print (colored("**DEBUG: cpu_here " +  str(threads_job) + " **", 'yellow'))

	## send for each sample
	print ("+ Calling fastqc for samples...")	
	with concurrent.futures.ThreadPoolExecutor(max_workers=int(max_workers_int)) as executor:
		commandsSent = { executor.submit(fastqc_caller.run_module_fastqc, outdir_dict[name], sorted( cluster["sample"].tolist() ), name, threads_job): name for name, cluster in sample_frame }
		
		for cmd2 in concurrent.futures.as_completed(commandsSent):
			details = commandsSent[cmd2]
			try:
				data = cmd2.result()
			except Exception as exc:
				print ('***ERROR:')
				print (cmd2)
				print('%r generated an exception: %s' % (details, exc))

	print ("+ FASTQC for samples has finished...")	
	
	## functions.timestamp
	start_time_partial = functions.timestamp(start_time_partial)

	if (options.skip_report):
		print ("+ No report generation...")
	else:
		print ("\n+ Generating a report using MultiQC module.")
		outdir_report = functions.create_subfolder("report", outdir)

		## get subdirs generated and call multiQC report module
		givenList = []
		print ("+ Detail information for each sample could be identified in separate folders:")
		
		## call multiQC report module
		givenList = [ v for v in outdir_dict.values() ]
		my_outdir_list = set(givenList)

		## debug message
		if (Debug):
			print (colored("\n**DEBUG: my_outdir_list for multiqc report **", 'yellow'))
			print (my_outdir_list)
			print ("\n")
		
		fastqc_report = functions.create_subfolder("FASTQC", outdir_report)
		fastqc_final_report = functions.create_subfolder(name_analysis, fastqc_report)
		multiQC_report.multiQC_module_call(my_outdir_list, "FASTQC", fastqc_final_report,"")
		print ('\n+ A summary HTML report of each sample is generated in folder: %s' %fastqc_final_report)

	print ("\n*************** Finish *******************")
	start_time_partial = functions.timestamp(start_time_total)

	print ("+ Exiting qc module.")
	exit()

################################################
def BUSCO_check(input_dir, outdir, options, start_time_total, mode):

	functions.boxymcboxface("BUSCO Assembly Quality check")

	## absolute path for in & out
	database_folder = os.path.abspath(options.database)

	## get files and get dir for each sample according to mode
	if mode == 'genome':
		pd_samples_retrieved = sampleParser.get_files(options, input_dir, "assembly", ["fna"])

		if not options.project:
			outdir = functions.create_subfolder("assembly_qc", outdir)

		if options.debug:
			print ("** DEBUG: pd_samples_retrieved")
			print (pd_samples_retrieved)
		
		BUSCO_outdir_dict = functions.outdir_project(outdir, options.project, pd_samples_retrieved, "assemble_qc")

	elif mode == 'proteins':
		pd_samples_retrieved = sampleParser.get_files(options, outdir, "annot", ["faa"]) ##

		if not options.project:
			outdir = functions.create_subfolder("annot_qc", outdir)

		if options.debug:
			print ("** DEBUG: pd_samples_retrieved")
			print (pd_samples_retrieved)
			
		BUSCO_outdir_dict = functions.outdir_project(outdir, options.project, pd_samples_retrieved, "annot_qc")

	## add column to dataframe
	pd_samples_retrieved['busco_folder'] = ""
	for index, row in pd_samples_retrieved.iterrows():
		pd_samples_retrieved.at[index, 'busco_folder'] = BUSCO_outdir_dict[ row['name'] ] 

	## debug message
	if (options.debug):
		print (colored("**DEBUG: df_samples_busco **", 'yellow'))
		print (pd_samples_retrieved)
		
		print (colored("**DEBUG: BUSCO_outdir_dict **", 'yellow'))
		print (BUSCO_outdir_dict)

	## Check each using BUSCO
	database_folder = os.path.abspath(options.database)
	BUSCO_Database = database_folder + '/BUSCO'
	if not os.path.exists(BUSCO_Database):
		functions.create_folder(BUSCO_Database)

	## call
	(dataFrame_results, stats_results) = BUSCO_call(options.BUSCO_dbs, pd_samples_retrieved, BUSCO_Database, options.threads, mode)
	
	## debug message
	if (options.debug):
		print (colored("**DEBUG: dataFrame_results **", 'yellow'))
		pd.set_option('display.max_colwidth', None)
		pd.set_option('display.max_columns', None)
		print (dataFrame_results)
	
	## functions.timestamp
	print ("+ Quality control of all samples finished: ")
	start_time_partial = functions.timestamp(start_time_total)
	
	## multiqc report plot
	if (options.skip_report):
		print ("+ No report generation...")
	else:
		print ("\n+ Generating a report BUSCO plot.")
		outdir_report = functions.create_subfolder("report", outdir)

		## get subdirs generated and call multiQC report module
		givenList = []
		print ("+ Detail information for each sample could be identified in separate folders.")
		
		## name folder according to mode
		if mode == 'genome':
			BUSCO_report = functions.create_subfolder("BUSCO_assembly", outdir_report)
		elif mode == 'proteins':
			BUSCO_report = functions.create_subfolder("BUSCO_annot", outdir_report)

		## generate plots
		print ("+ Generate summarizing plots...")
		BUSCO_plots(dataFrame_results, BUSCO_report, options.threads)	
		print ('\n+ Check quality plots in folder: %s' %BUSCO_report)

		##	TODO 
		##	Parse BUSCO statistics in dataframe (stats_results) for discarding samples if necessary
		##	given a cutoff, discard or advise to discard some samples

		### print statistics
		stats_results.to_csv(BUSCO_report + "/BUSCO_stats.csv")
		name_excel = BUSCO_report + "/BUSCO_stats.xlsx"
		writer = pd.ExcelWriter(name_excel, engine='xlsxwriter')
		stats_results.to_excel(writer, sheet_name="BUSCO statistics")	
		writer.save()
		
		print ('\n+ Check quality statistics in folder: %s' %BUSCO_report)
	
	return(dataFrame_results)

################################################
def BUSCO_call(datasets, pd_samples, database_folder, threads, mode):
	## 
	## argument mode = proteins or genome
	##
	
	## get datasets
	print ("+ Check folder provided as database for available BUSCO datasets...")
	BUSCO_datasets = BUSCO_caller.BUSCO_retrieve_sets(datasets, database_folder)
	
	## BUSCO needs to chdir to output folder
	path_here = os.getcwd()
	
	print ("+ Checking quality for each sample retrieved...")
	## optimize threads: No need to optimize. There is a problem with the working dir of BUSCO and we need to change every time
	# We can use a with statement to ensure threads are cleaned up promptly
	with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor: ## need to do 1 by one as there is a problem with the working directory
		for DataSet in BUSCO_datasets:
			## send for each sample
			commandsSent = { executor.submit( BUSCO_runner, row['name'], DataSet, row['sample'], BUSCO_datasets[DataSet], row['busco_folder'], threads, mode): name for name, row in pd_samples.iterrows() }

			for cmd2 in concurrent.futures.as_completed(commandsSent):
				details = commandsSent[cmd2]
				try:
					data = cmd2.result()
				except Exception as exc:
					print ('***ERROR:')
					print (cmd2)
					print('%r generated an exception: %s' % (details, exc))
			
			print ("+ Jobs finished for dataset %s\n+ Collecting information..." %DataSet)

	print ("Finish here...")
	os.chdir(path_here)
	
	## init dataframe
	short_summary = pd.DataFrame(columns=('sample', 'dirname', 'name', 'ext', 'tag', 'busco_folder', 'busco_dataset', 'busco_summary', 'busco_results'))
	stats_summary = pd.DataFrame()

	## generate results
	for DataSet in BUSCO_datasets:
		for index, row in pd_samples.iterrows():
			#my_BUSCO_results_folder = row['busco_folder'] + '/run_' + DataSet
			my_BUSCO_results_folder = row['busco_folder'] + '/' + row['name'] + '/run_' + DataSet
			my_short_txt = 	my_BUSCO_results_folder + '/short_summary_' + DataSet + '.txt'
			
			if os.path.isfile(my_short_txt):
				short_summary.loc[len(short_summary)] = [ row['sample'], row['dirname'], row['name'], row['ext'], row['tag'], row['busco_folder'], DataSet, my_short_txt, my_BUSCO_results_folder ]
				my_stats = BUSCO_caller.BUSCO_stats(my_short_txt, row['name'], DataSet)
				stats_summary = pd.concat([stats_summary, my_stats])
		
	return (short_summary, stats_summary)

################################################
def BUSCO_runner(sample_name, DataSet, sample, dataset_path, output, threads, mode):
	#file_name = os.path.basename(sample)
	sample_name_dir = functions.create_subfolder(sample_name, output) ## to check in detached mode

	## run busco	
	code = BUSCO_caller.BUSCO_run( dataset_path, sample, threads, sample_name_dir, DataSet, mode)
	
	## retry it just in case
	if (code == 'FAIL'):
		BUSCO_caller.BUSCO_run( dataset_path, sample, threads, sample_name_dir, DataSet, mode)
	else:
		return ()

################################################
def BUSCO_plots(dataFrame_results, outdir, threads):

	## DataFrame columns ('sample', 'dirname', 'name', 'ext', 'tag', 'busco_folder', 'busco_dataset', 'busco_summary', 'busco_results'))
	list_datasets = set(dataFrame_results['busco_dataset'].tolist())
	list_samples = set(dataFrame_results['name'].tolist())

	plot_folder = functions.create_subfolder('BUSCO_plots', outdir)
	outdir_busco_plot = []
	
	print ("+ Get results for all samples summarized by dataset:")
	for dataset in list_datasets:
		print ("\t+ Get results for: ", dataset)
		plot_folder_dataset = functions.create_subfolder(dataset, plot_folder)
		outdir_busco_plot.append(plot_folder_dataset)
	
		for index, row in dataFrame_results.iterrows():
			if (dataset == row['busco_dataset']):
				shutil.copy(row['busco_summary'], plot_folder_dataset + '/short_summary_' + dataset + '_' + row['name'] + '.txt')
		
	print ("+ Get results for summarized by sample:")
	for sample in list_samples:
		print ("\t+ Get results for: ", sample)
		plot_folder_sample = functions.create_subfolder(sample, plot_folder)
		outdir_busco_plot.append(plot_folder_sample)

		for index, row in dataFrame_results.iterrows():
			if (sample == row['name']):
				shutil.copy(row['busco_summary'], plot_folder_sample + '/short_summary_' + row['busco_dataset'] + '_' + sample + '.txt')
	
	print ("+ Generate plots for each subset")
	
	## optimize threads
	threads_job = 1  ## threads optimization
	max_workers_int = threads

	# We can use a with statement to ensure threads are cleaned up promptly
	with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers_int) as executor: ## need to do 1 by one as there is a problem with the working directory
		## send for each sample
		commandsSent = { executor.submit( BUSCO_caller.BUSCO_plot , plot): plot for plot in outdir_busco_plot }

		for cmd2 in concurrent.futures.as_completed(commandsSent):
			details = commandsSent[cmd2]
			try:
				data = cmd2.result()
			except Exception as exc:
				print ('***ERROR:')
				print (cmd2)
				print('%r generated an exception: %s' % (details, exc))
			
	print ("+ All plots generated...")
	print ("+ Check results under folders in : ", plot_folder)		




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
"""
Calls ARIBA_ software for several databases and custom dataset.
"""
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
from io import open
from termcolor import colored
import pandas as pd
from ete3 import Tree

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.modules import citation

############################################################### 
def get_ARIBA_dbs(list_dbs):
	
	

	## General
	dbs = [
		"argannot", "card", "megares", 
		"plasmidfinder", "resfinder", 
		"srst2_argannot", "vfdb_core", 
		"vfdb_full", "virulencefinder"]
	
	conversion_dbs = {
		"ARG-ANNOT":"argannot", 
		"CARD":"card", 
		"MEGARes":"megares", 
		"PlasmidFinder":"plasmidfinder", 
		"ResFinder":"resfinder", 
		"srst2":"srst2_argannot", 
		"VFDB":"vfdb_full", 
		"VirulenceFinder":"virulencefinder"
	}
	
	db2returns = []
	for i in list_dbs:
		if (conversion_dbs[i] in dbs):
			db2returns.append(conversion_dbs[i])
	return (db2returns)
	
############################################################### 
def help_ARIBA():

	dict_ariba = citation.ariba_citation()
	## to do: finish filling information for different databases
	print ("")
	functions.print_sepLine("*", 50, False)
	print ("CARD:")
	print ("The Comprehensive Antibiotic Resistance Database (CARD) is a rigorously curated collection of characterized, peer-reviewed  resistance determinants and associated antibiotics, organized by the Antibiotic Resistance Ontology (ARO) and AMR gene detection models.")
	print ('Citation:', dict_ariba['CARD'])
	print ("")	
	functions.print_sepLine("*", 50, False)
	print ("VFDB:")
	print ("The virulence factor database (VFDB) is an integrated and comprehensive online resource for curating information about virulence factors of bacterial pathogens. Since its inception in 2004, VFDB has been dedicated to providing up-to-date knowledge of VFs from various medically significant bacterial pathogens.")
	print ('Citation:', dict_ariba['VFDB'])
	print ("")	
	functions.print_sepLine("*", 50, False)
	print ("ARG-ANNOT:\n")
	print ("...")
	print ('Citation:', dict_ariba['ARG-ANNOT'])
	print ("")	
	functions.print_sepLine("*", 50, False)
	print ("MEGARes:")
	print ("The MEGARes database contains sequence data for approximately 4,000 hand-curated antimicrobial resistance genes accompanied by an annotation structure that is optimized for use with high throughput sequencing.")
	print ('Citation:', dict_ariba['MEGARes'])
	print ("")	
	functions.print_sepLine("*", 50, False)
	print ("PlasmidFinder:\n")
	print ("...")
	print ('Citation:', dict_ariba['PlasmidFinder'])
	print ("")	
	functions.print_sepLine("*", 50, False)
	print ("ResFinder:\n")
	print ("The ResFinder database is a curated database of acquired resistance genes.")
	print ('Citation:', dict_ariba['ResFinder'])
	print ("")	
	functions.print_sepLine("*", 50, False)
	print ("srst2:")
	print ("...")
	print ('Citation:', dict_ariba['srst2'])
	print ("")
	functions.print_sepLine("*", 50, False)
	print ("")
	
############################################################### 
def download_ariba_databases(list_dbs, main_folder, Debug, threads):

	"""Download ARIBA_ databases.
	
	Using ARIBA software this function retrieves desired databases and prepare them for later analysis.
	
	:param list_dbs: List of databases to download.
	:param main_folder: Absolute path to database folder.
	:param Debug: True/false for printing developer messages
	:param threads: Number of CPUs to use.
	
	:type list_dbs: string 
	:type main_folder: string
	:type Debug: Boolean
	:type threads: integer
	
	 .. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.create_subfolder`
		
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`
		
		- :func:`BacterialTyper.scripts.ariba_caller.get_ARIBA_dbs`
	
		- :func:`BacterialTyper.scripts.ariba_caller.ariba_getref`		
		
	 
	.. include:: ../../links.inc
	"""

	print("\n\n+ Download databases for Antimicrobial Resistance Identification By Assembly (ARIBA).")
	ariba_folder = functions.create_subfolder("ARIBA", main_folder)

	## print ARIBA databases: 
	print ("+ Available databases:")
	dbs = get_ARIBA_dbs(list_dbs)
	
	for db_set in dbs:

		functions.print_sepLine("-",30, False)
		print (colored("+ " + db_set,'yellow'))
		
		## prepare folders
		folder_set = functions.create_subfolder(db_set, ariba_folder)
		outdir_prepare_ref = folder_set + '_prepareref'

		## stamp time file
		filename_stamp_prepare = outdir_prepare_ref + '/.success'
	
		## check if previously done
		if os.path.isfile(filename_stamp_prepare):
			stamp =	functions.read_time_stamp(filename_stamp_prepare)
			print ("\t+ Database is downloaded in folder: ", folder_set)
			print ("\t+ Data is available and indexed in folder: ", outdir_prepare_ref)
			print (colored("\tDatabase was previously downloaded and prepared on: %s" %stamp, 'yellow'))
		
			## [TODO]: Check if necessary to download again after several months/days
			return_ariba_getref = 'OK'
		else:
			return_ariba_getref = ariba_getref(db_set, folder_set, Debug, threads)
		
		if (return_ariba_getref == 'OK'):
			print()
		else:
			print (colored("** ARIBA getref failed or generated a warning for " + db_set, 'red'))


##########
def check_db_indexed(folder, option):
	"""Check if ARIBA_ database is indexed.
	
	In the given folder it looks for '00.info.txt' file.
	
	:param folder: Absolute path to database folder.
	:param option: Whether to print more information messages or not [Yes/No]. 
	
	:type folder: string 
	:type option: string
	:returns: Boolean True/False.
	 
	 
	 .. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.read_time_stamp`
	 
	.. include:: ../../links.inc
	"""
	
	path_basename = folder.split('/')
	if os.path.isfile(folder + '00.info.txt'):
		if os.path.isfile(folder + '.success'):
			stamp =	functions.read_time_stamp(folder + '.success')
			print (colored("\tA previous command generated results on: %s [%s]" %(stamp, path_basename[-2]), 'yellow'))
			return True
	else:
		if (option == 'YES'):
			print (colored("\t- ARIBA database: " + path_basename + " [ ERROR ]", 'red'))
		return False
	
##########
def ariba_summary(out, infileList, options):
	"""Create ARIBA summary
	
	This function calls ARIBA_ summary and generates a summary of multiple ARIBA report files. It also creates Phandango_ files.
	
	:param out: Prefix of output files
	:param infileList: Files to be summarised
	:param options: Additional options for ariba summary. See details below. Provide them within quotes.
	
	:type out: string
	:type infileList: list
	:type options: string
	
	:returns: functions.system_call message
	
	.. note:: Additional options for ariba_summary (as specified by ARIBA_)
	
		--cluster_cols col_ids			Comma separated list of cluster columns to include. Choose from: assembled, match, ref_seq, pct_id, ctg_cov, known_var, novel_var [match]

		--no_tree             Do not make phandango tree

		--min_id FLOAT        Minimum percent identity cutoff to count as assembled [90]

		--only_clusters Cluster_names			Only report data for the given comma-separated list of cluster names, eg: cluster1,cluster2,cluster42
		
		--v_groups      Show a group column for each group of variants
		
		--known_variants      Report all known variants
		
		--novel_variants      Report all novel variants
	
		--col_filter string			Choose whether columns where all values are "no" or "NA" are removed [yes/no]

		--row_filter string			Choose whether rows where all values are "no" or "NA" are removed [yes/no]


	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.system_call`
	
	"""
	
	logFile = out + '.log'
	infileList_string = " ".join(infileList)
	cmd_summary = 'ariba summary %s %s %s 2> %s' %(options, out, infileList_string, logFile)
	
	return(functions.system_call(cmd_summary))

##########
def ariba_getref(database, outdir, Debug, threads):
	######################################################################################
	## usage: ariba getref [options] <db> <outprefix>
	######################################################################################
	## Download reference data from one of a few supported public resources
	## positional arguments:
	##	DB name            Database to download. Must be one of: argannot card megares plasmidfinder resfinder srst2_argannot vfdb_core vfdb_full virulencefinder
	##  outprefix          Prefix of output filenames
	######################################################################################

	## where database is one of: 
	##	argannot, card, megares, plasmidfinder, resfinder,
	##	srst2_argannot, vfdb_core, vfdb_full, virulencefinder.

	## folders
	outdir_name = outdir + '/' + database
	outdir_prepare_ref = outdir + '_prepareref'

	## download information in database folder provided by config
	print ("\t+ Retrieve information from database: " + database)

	## check if previously downloaded and succeeded
	filename_stamp = outdir + '/.success'

	if os.path.isfile(filename_stamp):
		stamp =	functions.read_time_stamp(filename_stamp)
		print (colored("\tA previous command generated results on: %s [%s]" %(stamp, name), 'yellow'))
		download_ariba_cmd = 'OK'
	else:
		cmd_getref = 'ariba getref %s %s' %(database, outdir_name)
		download_ariba_cmd = functions.system_call(cmd_getref)
	
	if (download_ariba_cmd == 'OK'):
		stamp =	functions.print_time_stamp(filename_stamp)
		## debug message
		if (Debug):
			print (colored("**DEBUG: ariba getref %s succeed " %database + "**", 'yellow'))

	else: 
		## rise error & exit
		print (colored("***ERROR: ariba getref %s failed " %database + " **",'red'))
		return('FAIL')

	## debug message
	if (Debug):
		print (colored("**DEBUG: Run ariba prepareref %s " %database + "**", 'yellow'))

	## check if previously prepareref and succeeded
	filename_stamp_prepare = outdir_prepare_ref + '/.success'
	if os.path.isfile(filename_stamp_prepare):
		stamp =	functions.read_time_stamp(filename_stamp_prepare)
		print (colored("\tA previous command generated results on: %s [%s]" %(stamp, name), 'yellow'))
	
	else:
		## get information
		list_files = os.listdir(outdir)
		fasta = ""
		metadata = ""
		for f in list_files:
			if f.endswith('tsv'):
				metadata = outdir + '/' + f
			elif f.endswith('fa'):
				fasta = outdir + '/' + f
	
		code = ariba_prepareref(fasta, metadata, outdir_prepare_ref, threads)
		
		if (code == 'OK'):
			filename_stamp = outdir_prepare_ref + '/.success'

		functions.print_time_stamp(filename_stamp_prepare)

	return()		
	
##########
def ariba_prepareref(fasta, metadata, outfolder, threads):
	## prepareref
	cmd_prepareref = 'ariba prepareref -f %s -m %s %s --threads %s' %(fasta, metadata, outfolder, threads)
	return(functions.system_call(cmd_prepareref))
	
	######################################################################################
	## usage: ariba prepareref [options] <outdir>
	######################################################################################
	## Prepare reference data for running the pipeline with "ariba run"
	## positional arguments:
	##   outdir                Output directory (must not already exist)
	## optional arguments:
	## -h, --help            show this help message and exit
	## input files options:
	##   -f FILENAME, --fasta FILENAME
	## 		REQUIRED. Name of fasta file. Can be used more than once if your sequences are spread over more than on file
	##   -m FILENAME, --metadata FILENAME
	##      Name of tsv file of metadata about the input sequences. Can be used more than once if your metadata is spread over more than one file. Incompatible with --all_coding
	##   --all_coding {yes,no}
	##      Use this if you only have a fasta of presence absence sequences as input, and no metadata. Use "yes" if all sequences are coding, or "no" if they are all non- coding. Incompatible with -m/--metadata
	## 
	## cd-hit options:
	##   --no_cdhit Do not run cd-hit. Each input sequence is put into its own "cluster". Incompatible with --cdhit_clusters.
	##   --cdhit_clusters FILENAME --> File specifying how the sequences should be clustered. Will be used instead of running cdhit. Format is one cluster per line. Sequence names separated by whitespace. Incompatible with --no_cdhit
	##   --cdhit_min_id FLOAT  Sequence identity threshold (cd-hit option -c) [0.9]
  	## 	 --cdhit_min_length FLOAT Length difference cutoff (cd-hit option -s) [0.0]
	##   --cdhit_max_memory INT Memory limit in MB (cd-hit option -M) [None]. Use 0 for unlimited.
	##
	## other options:
	##   --min_gene_length INT -->  Minimum allowed length in nucleotides of reference genes [6]
	##   --max_gene_length INT -->  Maximum allowed length in nucleotides of reference genes [10000]
	##   --genetic_code INT   -->   Number of genetic code to use. Currently supported 1,4,11 [11]
	##   --force               Overwrite output directory, if it already exists
	##   --threads INT         Number of threads (currently only applies to cdhit)[1]
	##  --verbose             Be verbose
	######################################################################################

##########
def ariba_expandflag(input_file, output_file):
	######################################################################################
	## usage: ariba expandflag infile.tsv outfile.tsv
	######################################################################################	
	#	Expands the flag column in a report file from number to comma-separated list
	#	of flag bits
	#
	#	positional arguments:
	#	  infile      Name of input report TSV file
	#	  outfile     Name of output report TSV file
	######################################################################################
	
	## download information in database folder provided by config
	#print ("+ Call ariba module 'expandflag' to add additional information to each entry.")
	
	## Somehow when I do expandflag of report.tsv file I get spaces convert to tabs and everything is 
	## distorted. I had to make this 
	
	data = pd.read_csv(input_file, header=0, sep='\t')
	list_data2 = set(list(data['flag']))
	tmp_name = os.path.splitext(input_file)[0]
	
	## print only flag
	flag_file = tmp_name + '-tmp.tsv'
	flag_file_hd = open(flag_file, "w")	
	flag_file_hd.write('#flag_name\tflag')
	flag_file_hd.write('\n')
	
	for line in zip(list_data2, list_data2):
		string = ('flag_{}\t{}'.format(*line))
		flag_file_hd.write(string)
		flag_file_hd.write('\n')
	flag_file_hd.close()
	
	## generate description
	flag_file_out = tmp_name + '-description.tsv'
	cmd = 'ariba expandflag "%s" %s' %(flag_file, flag_file_out)
	functions.system_call(cmd)
	
	os.remove(flag_file)	
	return(flag_file_out)

##########
def ariba_pubmlstget(species, outdir):
	######################################################################################
	## usage: ariba pubmlstget [options] <"species in quotes"> <output_directory>
	######################################################################################	
	## Download typing scheme for a given species from PubMLST, and make an ARIBA db
	## positional arguments:
	##  species     Species to download. Put it in quotes
	##	outdir      Name of output directory to be made (must not already exist)
	######################################################################################
	
	## download information in database folder provided by config
	print ("+ Call ariba module 'pubmlstget' to retrieve MLST information.")
	functions.create_folder(outdir)
	cmd = 'ariba pubmlstget "%s" %s' %(species, outdir)
	return(functions.system_call(cmd))
	
##########
def ariba_run(database, files, outdir, threads, threshold):
	"""ARIBA search call
	
	Given a database, it generates an ARIBA search of the reads provided. Results would be later processed using :func:`BacterialTyper.scripts.virulence_resistance.check_results` and :func:`BacterialTyper.scripts.virulence_resistance.results_parser`
	
	:param database: Folder containing ARIBA database previously indexed.
	:param files: List of files with absolute path to fastq reads (R1 & R2).
	:param outdir: Absolute path to folder, must not exist.
	:param threads: Number of CPUs to use.
	:param threshold: If proportion of gene assembled (regardless of into how many contigs) is at least this value then the flag gene_assembled is set [Default: 0.90].
	
	:type database: string
	:type files: list
	:type outdir: string
	:type threads: integer
	:type threshold: float [0-1]
	
	:return: Message: OK/FAIL.
	:rtype: string
	
	:warnings: No single mode available, only paired-end.
	
	:warnings: Output directory must not exist.
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.functions.system_call`
		
		- :func:`BacterialTyper.scripts.functions.print_time_stamp`
		
		- :func:`BacterialTyper.scripts.virulence_resistance.check_results` 
		
		- :func:`BacterialTyper.scripts.virulence_resistance.results_parser`
	
	
	"""
	
	######################################################################################
	##	usage: ariba run [options] <prepareref_dir> <reads1.fq> <reads2.fq> <outdir>
	##	Runs the local assembly pipeline. Input is dir made by prepareref, and paired reads
	##
	## positional arguments:
	##  prepareref_dir        Name of output directory when "ariba prepareref" was run
	##  reads_1               Name of fwd reads fastq file
	##  reads_2               Name of rev reads fastq file
	##  outdir                Output directory (must not already exist)
	## optional arguments:
	##  -h, --help            show this help message and exit
	## nucmer options:
	##  --nucmer_min_id INT   Minimum alignment identity (delta-filter -i) [90]
	##  --nucmer_min_len INT  Minimum alignment length (delta-filter -i) [20]
	##  --nucmer_breaklen INT Value to use for -breaklen when running nucmer [200]
	## Assembly options:
	##  --assembler {fermilite,spades} Assembler to use
	##  --assembly_cov INT    Target read coverage when sampling reads for assembly [50]
	##  --min_scaff_depth INT Minimum number of read pairs needed as evidence for scaffold link between two contigs [10]
	##  --spades_mode {wgs,sc,rna}  If using Spades assembler, either use default WGS mode, Single Cell mode (`spades.py --sc`) or RNA mode (`spades.py --rna`). Use SC or RNA mode if your input is from a viral sequencing with very uneven and deep                        coverage. Set `--assembly_cov` to some high value if using SC or RNA mode
	##  --spades_options SPADES_OPTIONS: Extra options to pass to Spades assembler. Sensible default options will be picked based on `--spades_mode` argument.  Anything set here will replace the defaults completely
	## Other options:
	##  --threads INT         Experimental. Number of threads. Will run clusters in parallel, but not minimap (yet) [1]
	##  --assembled_threshold FLOAT (between 0 and 1)  If proportion of gene assembled (regardless of into how many contigs) is at least this value then the flag gene_assembled is set [0.95]
	##  --gene_nt_extend INT  Max number of nucleotides to extend ends of gene matches to look for start/stop codons [30]
	##  --unique_threshold FLOAT (between 0 and 1) If proportion of bases in gene assembled more than once is <= this value, then the flag unique_contig is set [0.03]
	##  --force               Overwrite output directory, if it already exists
	##  --noclean             Do not clean up intermediate files
	##  --tmp_dir TMP_DIR     Existing directory in which to create a temporary directory used for local assemblies
	##  --verbose             Be verbose
	######################################################################################

	# outdir must not exist.
	logFile = outdir + '_run.log'
	
	######################################################
	## default options: might be interesting to change?
	##  --assembled_threshold 0.95
	##  --nucmer_min_id INT	90
	##  --nucmer_min_len INT 20
	######################################################
	
	if (len(files) == 2):
		cmd = 'ariba run --assembled_threshold %s %s %s %s %s --threads %s 2> %s' %(threshold, database, files[0], files[1], outdir, threads, logFile)
	else:
	
		## [TODO]
		print (colored("\n\n***** TODO: Check what to do with single end when calling ARIBA *****\n\n", 'red'))
		print (colored("\n\n***** No implementation yet *****\n\n", 'red'))
		exit()

	##
	code = functions.system_call(cmd)

	## make stamp time
	if (code == 'OK'): 
		filename_stamp = outdir + '/.success'
		functions.print_time_stamp(filename_stamp)
		return ('OK')
	else:
		return ('FAIL')
	
#############################################################
def ariba_summary_all(outfile, dict_files):
	"""Create final ARIBA summary 
	
	Generates a temporal summary of all the reports generated by ARIBA (using :func:`BacterialTyper.scripts.ariba_caller.ariba_summary`). The result is a csv file and newick tree for later Phandango_ visualization.
	
	These files contained for each entry a ".match" string appended. Using :func:`BacterialTyper.scripts.ariba_caller.fix_ariba_summary` and :func:`BacterialTyper.scripts.ariba_caller.fix_phandando_tree`, we would removed this ".match" and prepare files 
	
	Final result files are a csv file (containing code colors) and newick tree for later Phandango_ visualization.
	
	:param outfile:
	:param dict_files: 
	
	:type outfile:
	:type dict_files: 
	
	:return: Output file
	:rtype: string
		
	
	.. seealso:: This function depends on other BacterialTyper functions called:
	
		- :func:`BacterialTyper.scripts.ariba_caller.ariba_summary`
		
		- :func:`BacterialTyper.scripts.ariba_caller.fix_ariba_summary` 
		
		- :func:`BacterialTyper.scripts.ariba_caller.fix_phandango_tree`
	
	"""
	
	## Create tmp list of files 
	file_list = []
	for files in dict_files:
		file_list.append(dict_files[files])
	
	## call ariba_summary
	outfile_tmp = outfile + '_tmp'
	options = ''
	ariba_summary(outfile_tmp, file_list, options)

	## fix output
	fix_ariba_summary(outfile_tmp + '.csv', outfile + '.csv', dict_files)
	fix_ariba_summary(outfile_tmp + '.phandango.csv', outfile + '.phandango.csv', dict_files)
	fix_phangando_tree(outfile_tmp + '.phandango.tre', outfile + '.phandango.tree', dict_files)
	
	if os.path.isfile(outfile_tmp + '.csv'):
		os.remove(outfile_tmp + '.csv')
	else:
		return('NaN')

	if os.path.isfile(outfile_tmp + '.phandango.csv'):
		os.remove(outfile_tmp + '.phandango.csv')

	if os.path.isfile(outfile_tmp + '.phandango.tree'):
		os.remove(outfile_tmp + '.phandango.tree')


	return(outfile + '.csv')
	
#############################################################
def fix_ariba_summary(csv_file, outfile, dict_files):

	if not os.path.isfile(csv_file):
		## summary file not generated
		print (colored("*** ERROR: summary file (%s) not generated ***" %csv_file, 'red'))
		return()

	summary_data = pd.read_csv(csv_file, header=0, sep=',')
	## parse results
	cluster = ['name']
	for column in summary_data:
		if (column == 'name'):
			continue
		cluster_name = column.replace(".match", "")
		cluster.append(cluster_name)

	## format summary data
	summary_data.columns = cluster 
	for index, row in summary_data.iterrows():
		for key, value in dict_files.items():
			if (value == row['name'] ):
				summary_data.loc[index]['name'] = key
	
	summary_data = summary_data.set_index('name')
	summary_data.to_csv(outfile)
	return()
	
#############################################################
def fix_phangando_tree(tree_file, outfile_name, dict_files): 
	
	if not os.path.isfile(tree_file):
		## summary file not generated
		print (colored("*** ERROR: phandango tree file (%s) not generated ***" %tree_file, 'red'))
		return()

	t = Tree(tree_file)
	for leaf in t:
		for key, value in dict_files.items():
			if (value == leaf.name):
				leaf.name = key

	# We can also write into a file
	t.write(format=1, outfile=outfile_name)
	return()


######
def	help_options():
	print ("\nUSAGE: python %s file name xx threads path\n"  %os.path.realpath(__file__))
	print ("*** if paired-end data provide a csv string for file argument ***\n\n")

######
def main():
	## this code runs when call as a single script

  	## control if options provided or help
	if len(sys.argv) > 1:
		print ("")
	else:
		help_options()
		exit()    	

	help_ARIBA()
		
######
if __name__== "__main__":
	main()




#!/usr/bin/env python3
#################################################################
## Jose F. Sanchez                                             ##
## Copyright (C) 2019-2020 Lauro Sumoy Lab, IGTP, Spain        ##
#################################################################
"""
Parses and prepares for phylogenetic analysis
"""
## useful imports
import time
import io
import os
import re
import sys
from sys import argv
from io import open
from termcolor import colored
import pandas as pd
from ete3 import Tree
from Bio import AlignIO
import numpy
import scipy.cluster.hierarchy as sch
from scipy.sparse.csgraph import minimum_spanning_tree
import networkx
import pylab

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config
from BacterialTyper.scripts import min_hash_caller

def get_snp_distance(aln_file, mode, countGaps, output, Debug):
    """
    
    """

    (D, labeltext) = snp_distance(aln_file, mode, countGaps, Debug)

    ### Write output
    labeltext_string = [str(x) for x in labeltext] ## generat strings, avoid if integers alone
    labeloutname = output + '.labels.txt'
    with open(labeloutname, 'w') as fp:
        fp.write("\n".join(labeltext_string))
        
    ## save matrix in csv format file
    numpy.set_printoptions(suppress=True)
    numpy.savetxt(output + ".csv", D, delimiter=",", fmt='%5.4g', header= ",".join(labeltext_string) )

    ## create minimum spanning tree
        ## https://docs.scipy.org/doc/scipy/reference/generated/scipy.cluster.hierarchy.linkage.html
    Y = sch.linkage(D, method='single', optimal_ordering=True)
    min_hash_caller.get_Newick_tree(Y, D, labeltext, output)
        
        ##D_network = networkx.convert_matrix.from_numpy_matrix(D)
        ##MST = networkx.minimum_spanning_tree(D_network)
        ##print (MST)
        ##fig3 = pylab.figure(figsize=(15, 10))
        ##networkx.draw_networkx(MST, with_labels=True, node_size = 45)
        ##fig3.savefig(output + '.pdf')
        
    return()

#################
def snp_distance(aln_file, mode, countGaps, Debug):
    """
    Calculate SNP difference between records in alignment
    
    :param aln_file: Alignment file.
    :param mode: Format of aln_file: [nexus,phylip,clustalw,fasta]
    :param countGaps: True/false for counting gaps as differences.
    :param Debug: True/false for debugging messages.
    
    :type aln_file: string
    :type mode: string 
    :type countGaps: bool
    :type Debug: bool
    
    :returns: Numpy matrix with SNP distance matrix and list of labels.
    
    """
    
    ## read alingment
    alignment = AlignIO.read(aln_file, mode)
    
    # build the distance matrix
    D = numpy.zeros([len(alignment), len(alignment)])

    labeltext = []
    
    for i, record in enumerate(alignment):
        for j, record2 in enumerate(alignment):
            snps_diff = 0
            if (record.id != record2.id):
                snps_diff = distance(record.seq, record2.seq, gaps=countGaps)
        
            if Debug:
                print (colored("** DEBUG: record.id, record2.id, snps_diff", 'yellow'))
                print ("%s vs %s : %i" %(record.id, record2.id, snps_diff))
            D[i][j] = snps_diff
            D[j][i] = snps_diff
        
        labeltext.append(record.id)
        
    ## Debug messages
    if Debug:
        print (colored("\n*** DEBUG: SNP distance matrix *****\n", 'red'))
        print ('D:')
        print (D)
        print ('labeltext:')
        print (labeltext)
        print ('Min similarity in matrix: {:.3f}', numpy.min(D)) ## use this to color accordingly

        
    return (D, labeltext)

#################
def distance(seq1, seq2, gaps=False):
    """
    Get distance between strings
    
    :param seq1: DNA sequence string 1
    :param seq2: DNA sequence string 2
    :param gaps: True/false for counting gaps as differences
    
    :type seq1: string
    :type seq1: string
    :type gaps: bool
    
    :returns: Integer of snps distance
    """
    
    ## check length
    len1 = len(seq1)
    len2 = len(seq2)
    if (len1 != len2):
        print (colored("** ERROR: Length of DNA strings are not the same...", 'red'))
        exit()
    
    ## compare
    snps = 0
    for pos in range (0, min(len1, len2)) :
        if seq1[pos] != seq2[pos]:
            if (seq1[pos] == '-' or seq2[pos]  == '-'):
                if (gaps):
                    snps += 1
            else:
                snps += 1
               
    return(snps)

#################
def ml_tree(folder, name, threads, output, Debug):
    """
    Create Maximum Likelihood tree reconstruction 
    
    We use IQ-Tree for the versatility and the ability to automatically set parameters. 
    
    :param folder: Snippy-core folder containing results.
    :param name: Name of the analysis.
    :param Debug: True/false for debugging messages
    
    :type folder: string 
    :type name: string
    :type Debug: bool 
    """
    iqtree_exe = set_config.get_exe('iqtree', Debug) 
    bootstrap_number = '1000'
    aln_file = os.path.join(folder, name + '.aln')
    output_log = os.path.join(output, 'iqtree.error.log')
    output_files = os.path.join(output, 'iqtree_' + name)
    
    iqtree_cmd = '%s -s %s -redo --threads-max %s --prefix %s -B %s 2> %s' %(iqtree_exe, aln_file, 
                                                                      threads, output_files, 
                                                                      bootstrap_number, output_log)
    code = functions.system_call(iqtree_cmd)
    
    if code == 'OK':
        return ()
    else:
        print ("Some error occurred...")
        return()

    ## raxml is available as conda package
    ## https://anaconda.org/bioconda/raxml
    ## other possibility is FastTree   
    
######
def help_options():
    print ("\nUSAGE: python %s file_aln format[nexus,phylip,clustalw,fasta]\n"  %os.path.realpath(__file__))

######
def main():
    ## this code runs when call as a single script

    ## control if options provided or help
    if len(sys.argv) > 1:
        print ("")
    else:
        help_options()
        exit()        

    get_snp_distance(argv[1], argv[2], countGaps=True, output="test_phylo", Debug=True)
        
######
if __name__== "__main__":
    main()
    
    
    
    
    
    
    
    
    
    
    
    
    



#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Calls card-trick module to parse CARD resistance information.
'''
## useful imports
import time
import io
import os
import re
import sys
import pandas as pd
from sys import argv
from io import open
from termcolor import colored
import card_trick

## import my modules
from BacterialTyper.scripts import functions
from BacterialTyper.config import set_config

##########
def get_info_CARD(IDs, term, dataF):
	## dataF contains CARD ontology: downloaded and parse using prepare_card_data
	## IDs is an input list
	## term is the type of search to do using card_trick
	
	# search for terms provided	
	matching_terms = card_trick.ontology_functions.search(IDs, dataF, term, False)
	return (matching_terms)

#####################
def prepare_card_data(database_folder):
	
	## create CARD folder
	abs_folder = os.path.abspath(database_folder)
	CARD_folder = functions.create_subfolder('CARD', abs_folder)
	
	## make stamp time
	filename_stamp = CARD_folder + '/.success'

	if os.path.isfile(filename_stamp):
		stamp =	functions.read_time_stamp(filename_stamp)
		print (colored("\tA previous command generated results on: %s [CARD Ontology Data]" %stamp, 'yellow'))

		## check time passed
		days_passed = functions.get_diff_time(stamp)
		print ("\t+ %s days ago" %days_passed)		
		if (days_passed > 30): ## download again
			print ("+ Downloading information again just to be sure...")
			download=True
		else:
			print ("\t+ No need to download data again.")
			download=False
	else:
		download=True

	###
	if download:
		## uptade database in a path
		aro_obo_file = card_trick.ontology_functions.update_ontology(CARD_folder, False)
	
		## get ontology and save it in csv
		return_frame = card_trick.ontology_functions.parse_ontology(aro_obo_file, False)
	
		### if success return folder name
		if not return_frame.empty:
			## success stamps
			filename_stamp = CARD_folder + '/.success'
			stamp =	functions.print_time_stamp(filename_stamp)	
		else:
			return (FAIL)

	## return folder name
	return(CARD_folder)


def dead_code():
	## card_prepareref
	rename_info = card_prepareref + '00.rename_info'
	
	## outfile
	outfile = card_prepareref + '00.info_dictionary'
	out_file_handle = open(outfile, 'w')	
	
	## get info
	lines = functions.readList_fromFile(rename_info)
	for l in lines:
		names = l.split('\t')
		## original name \t ariba_name
		out_file_handle.write(names[1].split('.')[0] + '\t' + names[0].split('.')[0] + '\n')

	out_file_handle.close()	
	return (outfile)
	




#!/usr/bin/env python3
##########################################################
## Jose F. Sanchez										##
## Copyright (C) 2019 Lauro Sumoy Lab, IGTP, Spain		##
##########################################################
'''
Prints an index of version numbers for the different packages and other softwares employed here.
'''

## useful imports
from termcolor import colored
import os
import sys

## import my modules
from BacterialTyper.config import set_config
from BacterialTyper.config import extern_progs
from BacterialTyper.scripts import functions


#from BacterialTyper.scripts import __version__ as ariba_version ## to include when distributed
pipeline_version = "0.0.2"

#########
def print_all():

	functions.print_sepLine("+", 50, False)
	print ('Python:')
	functions.print_sepLine("+", 50, False)
	print ('Python version:', str(sys.version))
	print ('\n')
	functions.print_sepLine("+", 50, False)
	print ('Python packages:')
	extern_progs.print_package_version()
	functions.print_sepLine("+", 50, False)
	print ('\n')

	functions.print_sepLine("+", 50, False)
	print ('External dependencies:\n')
	functions.print_sepLine("+", 50, False)
	extern_progs.print_dependencies()
	print ('\n')

	print ('Additional dependencies: databases, information, etc...')

	functions.print_sepLine("*", 50, False)
	print ("ARIBA databases version..")
	functions.print_sepLine("*", 50, False)
	print ("card -> ")
	print ("megares  ->")
	print ("plasmidfinder  ->")
	print ("resfinder  ->")
	print ("srst2_argannot  ->")
	print ("vfdb_core & vfdb_full  ->")
	print ("virulencefinder  ->")
	print ('\n')
	
	## kma databases
	## ...
	
#########
def only_us():
	print ("Version to be included when properly finished...\n\n")

	functions.print_sepLine("+", 50, False)
	print ('BacterialTyper version: ' + pipeline_version)
	functions.print_sepLine("+", 50, False)
	print ('\n')


#########
def run(options):
	functions.pipeline_header()
	functions.boxymcboxface("Version")

	if (options.option == 'all'):
		print (colored("\n+ BacterialTyper version:", 'yellow'))
		only_us()

		print (colored("\n+ Other softwares employed in the pipeline", 'yellow'))
		print_all()
		
	elif (options.option == 'only'):
		print (colored("\n+ BacterialTyper version:", 'yellow'))
		only_us()

	return()
